@@ -10,7 +10,7 @@ import uuid
 import unittest
 
 from xpra.util import envint
-from xpra.exit_codes import EXIT_STR
+from xpra.exit_codes import exit_str
 from xpra.os_util import load_binary_file, pollwait, OSX, POSIX
 from xpra.platform.paths import get_download_dir
 from unit.client.x11_client_test_util import X11ClientTestUtil, log
@@ -24,23 +24,25 @@ class X11ClientTest(X11ClientTestUtil):
         display = self.find_free_display()
         log("starting test server on %s", display)
         server_args = ["--start=xterm"] + list(server_args)
-        server = self.check_start_server(display, *server_args)
+        server = self.check_fast_start_server(display, *server_args)
         xvfb1, client1 = self.run_client(display, *client_args)
         r = pollwait(client1, CLIENT_TIMEOUT)
-        assert r is None, "client1 exited with code %s" % EXIT_STR.get(r, r)
+        if r is not None:
+            raise RuntimeError(f"client1 with args {client_args} exited with code {exit_str(r)}")
         xvfb2, client2 = self.run_client(display, *client_args)
         r = pollwait(client2, CLIENT_TIMEOUT)
-        assert r is None, "client2 exited with code %s" % EXIT_STR.get(r, r)
+        if r is not None:
+            raise RuntimeError(f"client2 with args {client_args} exited with code {exit_str(r)}")
         if disconnect:
             #starting a second client should disconnect the first when not sharing
             assert pollwait(client1, 2) is not None, "the first client should have been disconnected"
         #killing the Xvfb should kill the client
         xvfb1.terminate()
         xvfb2.terminate()
-        assert pollwait(xvfb1, CLIENT_TIMEOUT) is not None
-        assert pollwait(xvfb2, CLIENT_TIMEOUT) is not None
-        assert pollwait(client1, CLIENT_TIMEOUT) is not None
-        assert pollwait(client2, CLIENT_TIMEOUT) is not None
+        assert pollwait(xvfb1, CLIENT_TIMEOUT) is not None, "xvfb1 has not terminated"
+        assert pollwait(xvfb2, CLIENT_TIMEOUT) is not None, "xvfb2 has not terminated"
+        assert pollwait(client1, CLIENT_TIMEOUT) is not None, "client1 has not terminated"
+        assert pollwait(client2, CLIENT_TIMEOUT) is not None, "client2 has not terminated"
         server.terminate()
 
 
@@ -85,7 +87,7 @@ class X11ClientTest(X11ClientTestUtil):
         for server_depth in (16, 24, 30):
             server_display = self.find_free_display()
             log("depth=%i starting test server on %s", server_depth, server_display)
-            server = self.check_start_server(server_display,
+            server = self.check_fast_start_server(server_display,
                                             "--start=xterm", "--sync-xvfb=50",
                                             "--pixel-depth=%i" % server_depth)
             self._test_client_depth(server_display)
@@ -93,14 +95,19 @@ class X11ClientTest(X11ClientTestUtil):
 
     def do_test_control_send_file(self, data):
         f = self._temp_file(data)
+        client = xvfb = server = None
         try:
             display = self.find_free_display()
-            server = self.check_start_server(display)
+            server = self.check_fast_start_server(display, "--file-transfer=yes")
             xvfb, client = self.run_client(display, "--file-transfer=yes")
-            assert pollwait(client, CLIENT_TIMEOUT) is None
+            r = pollwait(client, CLIENT_TIMEOUT)
+            assert r is None, f"client terminated unexpectedly with code {exit_str(r)}"
             #send a file to this client:
             send_file_command = ["control", display, "send-file", f.name, "0", "*"]
-            send_file = self.run_xpra(send_file_command)
+            env = os.environ.copy()
+            import secrets
+            env["XPRA_USER_UUID"] = secrets.token_urlsafe(16)
+            send_file = self.run_xpra(send_file_command, env=env)
             assert pollwait(send_file, CLIENT_TIMEOUT)==0, "send-file command returncode is %s" % send_file.poll()
             time.sleep(1)
             #now verify the file can be found in the download directory
@@ -109,11 +116,13 @@ class X11ClientTest(X11ClientTestUtil):
             readback = load_binary_file(filename)
             assert readback==data, "file data corrupted"
             os.unlink(filename)
-            #cleanup:
-            client.terminate()
-            xvfb.terminate()
-            server.terminate()
         finally:
+            if client:
+                client.terminate()
+            if xvfb:
+                xvfb.terminate()
+            if server:
+                server.terminate()
             f.close()
 
     def test_control_send_file(self):
