@@ -1,13 +1,18 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2021-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
+from io import BytesIO
 from time import monotonic
+from PIL import Image
 
+from xpra.net import compression
+from xpra.util import envbool
 from xpra.codecs.image_wrapper import ImageWrapper
+from xpra.codecs.loader import load_codec
 
 N = 10
 CODECS = ("enc_rgb", "enc_pillow", "enc_spng", "enc_webp", "enc_jpeg", "enc_avif")
@@ -16,23 +21,20 @@ CODECS = ("enc_rgb", "enc_pillow", "enc_spng", "enc_webp", "enc_jpeg", "enc_avif
 options = {
     #"quality" : 10,
     #"grayscale" : True,
-    }
+}
 
 
 def main(fmt="png", files=()):
-    assert len(files)>0, "specify images to benchmark"
-    from xpra.net import compression
+    assert len(files)>0, "specify images to use for benchmark"
     compression.init_all()
-    from xpra.codecs.loader import load_codec, get_codec
     encoders = []
     for codec in CODECS:
-        load_codec(codec)
-        enc = get_codec(codec)
+        enc = load_codec(codec)
         if enc and (fmt=="all" or fmt in enc.get_encodings()):
             encoders.append(enc)
-
-    from PIL import Image
+    index = 0
     for f in files:
+        index += 1
         img = Image.open(f)
         if img.mode not in ("RGBA", "RGB"):
             img = img.convert("RGB")
@@ -40,7 +42,7 @@ def main(fmt="png", files=()):
         w, h = img.size
         rgb_data = img.tobytes("raw")
         stride = w * len(pixel_format)
-        print("%s : %s" % (f, img))
+        print(f"{f:40} : {img}")
         image = ImageWrapper(0, 0, w, h,
                              rgb_data, pixel_format, len(pixel_format)*8, stride,
                              planes=ImageWrapper.PACKED, thread_safe=True)
@@ -56,26 +58,32 @@ def main(fmt="png", files=()):
                     try:
                         r = enc.encode(encoding, image, options)
                     except Exception:
-                        print("error on %s %s" % (enc.get_type(), enc.encode))
+                        print(f"error on {enc.get_type()} : {enc.encode}")
                         raise
                     if not r:
-                        print("Error: no data for %s %s" % (enc.get_type(), enc.encode))
+                        print(f"Error: no data for {enc.get_type()} : {enc.encode}")
                         break
                     size += len(r[1])
                 if not r:
                     continue
                 end = monotonic()
                 cdata = r[1]
+                if envbool("SAVE", False):
+                    filename = f"./benchmark-{index}-{enc.get_type()}.{encoding.replace('/','-')}"
+                    bdata = getattr(cdata, "data", cdata)
+                    with open(filename, "wb") as fsave:
+                        fsave.write(bdata)
                 ratio = 100*len(cdata)/len(rgb_data)
-                print("%-10s %-10s   :    %10.1f MPixels/s    size=%8iKB    %3.1f%%" % (
-                    encoding, enc.get_type(), w*h*N/(end-start)/1024/1024, size*N/1024, ratio))
+                mps = w*h*N/(end-start)/1024/1024
+                sizek = size*N//1024
+                print(f"{encoding:10} {enc.get_type():10} {mps:10.1f} MPixels/s  {sizek:>8}KB  {ratio:3.1f}%")
                 #verify that the png data is valid using pillow:
                 if encoding not in ("rgb24", "rgb32", "avif"):
-                    from io import BytesIO
                     buf = BytesIO(cdata.data)
                     img = Image.open(buf)
                     #img.show()
 
+
 if __name__ == '__main__':
     assert len(sys.argv)>1
     files = sys.argv[1:]
