@@ -1,11 +1,12 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2022 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 from time import monotonic
+from typing import Dict, Any, Tuple
 
-from xpra.exit_codes import EXIT_INTERNAL_ERROR
+from xpra.exit_codes import ExitCode
 from xpra.platform.features import REINIT_WINDOWS
 from xpra.platform.gui import (
     get_vrefresh,
@@ -14,20 +15,17 @@ from xpra.platform.gui import (
     )
 from xpra.scripts.main import check_display
 from xpra.scripts.config import FALSE_OPTIONS
-from xpra.net.common import MAX_PACKET_SIZE
+from xpra.net.common import MAX_PACKET_SIZE, PacketType
 from xpra.common import (
     adjust_monitor_refresh_rate, get_refresh_rate_for_value,
-    FULL_INFO,
+    FULL_INFO, SYNC_ICC,
     )
-from xpra.client.scaling_parser import (
+from xpra.scaling_parser import (
     parse_scaling, scaleup_value, scaledown_value, fequ, r4cmp,
     MIN_SCALING, MAX_SCALING, SCALING_EMBARGO_TIME,
     )
-from xpra.util import (
-    envint, envbool, log_screen_sizes, flatten_dict, typedict,
-    XPRA_SCALING_NOTIFICATION_ID,
-    )
-from xpra.client.mixins.stub_client_mixin import StubClientMixin
+from xpra.util import envint, log_screen_sizes, flatten_dict, typedict, NotificationID
+from xpra.client.base.stub_client_mixin import StubClientMixin
 from xpra.log import Logger
 
 log = Logger("screen")
@@ -36,9 +34,6 @@ scalinglog = Logger("scaling")
 
 MONITOR_CHANGE_REINIT = envint("XPRA_MONITOR_CHANGE_REINIT")
 
-SYNC_ICC = envbool("XPRA_SYNC_ICC", True)
-
-
 
 class DisplayClient(StubClientMixin):
     """
@@ -58,7 +53,7 @@ class DisplayClient(StubClientMixin):
         self.refresh_rate = 0
         self.desktop_fullscreen = False
         self.desktop_scaling = False
-        self.screen_size_change_timer = None
+        self.screen_size_change_timer = 0
 
         self.server_desktop_size = None
         self.server_actual_desktop_size = None
@@ -89,7 +84,7 @@ class DisplayClient(StubClientMixin):
     def cleanup(self):
         ssct = self.screen_size_change_timer
         if ssct:
-            self.screen_size_change_timer = None
+            self.screen_size_change_timer = 0
             self.source_remove(ssct)
 
 
@@ -111,7 +106,7 @@ class DisplayClient(StubClientMixin):
 
     ######################################################################
     # hello:
-    def get_caps(self) -> dict:
+    def get_caps(self) -> Dict[str,Any]:
         caps = {
             "randr_notify"  : True,
             "show-desktop"  : True,
@@ -148,19 +143,20 @@ class DisplayClient(StubClientMixin):
                      "up" if scaled_up else "down", sinfo, root_w, root_h)
             log_screen_sizes(root_w, root_h, sss)
         else:
-            root_w, root_h = u_root_w, u_root_h
             sss = ss
         caps["screen_sizes"] = sss
         monitors = self.get_monitors_info()
         caps["monitors"] = adjust_monitor_refresh_rate(self.refresh_rate, monitors)
         caps.update(self.get_screen_caps())
+        if self.desktop_fullscreen:
+            caps["desktop-fullscreen"] = True
         caps.update(flatten_dict({
             "dpi"               : self.get_dpi_caps(),
             "screen-scaling"    : self.get_scaling_caps(),
             }))
         return caps
 
-    def get_dpi_caps(self) -> dict:
+    def get_dpi_caps(self) -> Dict[str,Any]:
         #command line (or config file) override supplied:
         caps = {}
         dpi = 0
@@ -186,7 +182,7 @@ class DisplayClient(StubClientMixin):
         log("get_dpi_caps()=%s", caps)
         return caps
 
-    def get_scaling_caps(self) -> dict:
+    def get_scaling_caps(self) -> Dict[str,Any]:
         return {
             "" : True,
             "enabled"   : self.xscale!=1 or self.yscale!=1,
@@ -196,7 +192,7 @@ class DisplayClient(StubClientMixin):
     def get_vrefresh(self):
         return get_vrefresh()
 
-    def get_screen_caps(self) -> dict:
+    def get_screen_caps(self) -> Dict[str,Any]:
         caps = {
             "vrefresh"      : self.get_vrefresh(),
             "cursor" : {
@@ -206,16 +202,11 @@ class DisplayClient(StubClientMixin):
         aa = get_antialias_info()
         if aa:
             caps["antialias"] = aa
-        if SYNC_ICC:
-            caps.update({
-            "icc"          : self.get_icc_info(),
-            "display-icc"  : self.get_display_icc_info(),
-            })
         return caps
 
     #this is the format we should be moving towards
     #with proper namespace:
-    #def get_info(self) -> dict:
+    #def get_info(self) -> Dict:
     #    sinfo = self.get_screen_caps()
     #    sinfo["scaling"] = self.get_scaling_caps()
     #    sinfo["dpi"] = self.get_dpi_caps()
@@ -257,7 +248,7 @@ class DisplayClient(StubClientMixin):
                               self.server_multi_monitors, self.server_monitors)
         return True
 
-    def process_ui_capabilities(self, c : typedict):
+    def process_ui_capabilities(self, c : typedict) -> None:
         self.server_is_desktop = c.boolget("shadow") or c.boolget("desktop") or c.boolget("monitor")
         log("process_ui_capabilities(%s) desktop=%s, monitor=%s", c, c.boolget("desktop"), c.boolget("monitor"))
         if not c.boolget("monitor"):
@@ -289,8 +280,22 @@ class DisplayClient(StubClientMixin):
         #now that we have the server's screen info, allow scale changes:
         self.scale_change_embargo = 0
         self.set_max_packet_size()
+        self.send_icc_data()
 
-    def set_max_packet_size(self):
+    def send_icc_data(self) -> None:
+        if SYNC_ICC and "configure-display" in self.server_packet_types:
+            #it is now safe to send the colourspace data if we have any:
+            icc = self.get_icc_info()
+            dicc = self.get_display_icc_info()
+            if icc or dicc:
+                self.send("configure-display", {
+                    "icc": {
+                        "global"    : icc,
+                        "display"   : dicc,
+                        }
+                    })
+
+    def set_max_packet_size(self) -> None:
         p = self._protocol
         if not p or p.TYPE!="xpra":
             return
@@ -305,11 +310,11 @@ class DisplayClient(StubClientMixin):
         if maxw<=0 or maxh<=0 or maxw>=32768 or maxh>=32768:
             message = "invalid maximum desktop size: %ix%i" % (maxw, maxh)
             log(message)
-            self.quit(EXIT_INTERNAL_ERROR)
+            self.quit(ExitCode.INTERNAL_ERROR)
             raise SystemExit(message)
         if maxw>=16384 or maxh>=16384:
             log.warn("Warning: the desktop size is extremely large: %ix%i", maxw, maxh)
-        #max packet size to accomodate
+        #max packet size to accommodate
         # a full screen RGBX (32 bits) uncompressed image
         # also with enough headroom for some metadata (4k)
         #we can't assume to have a real ClientConnection object:
@@ -321,21 +326,21 @@ class DisplayClient(StubClientMixin):
     def has_transparency(self) -> bool:
         return False
 
-    def get_icc_info(self) -> dict:
+    def get_icc_info(self) -> Dict[str,Any]:
         return get_icc_info()
 
-    def get_display_icc_info(self) -> dict:
+    def get_display_icc_info(self) -> Dict[str,Any]:
         return get_display_icc_info()
 
     def get_monitors_info(self):
         return {}
 
-    def _process_show_desktop(self, packet):
+    def _process_show_desktop(self, packet : PacketType) -> None:
         show = packet[1]
         log("calling %s(%s)", show_desktop, show)
         show_desktop(show)
 
-    def _process_desktop_size(self, packet):
+    def _process_desktop_size(self, packet : PacketType) -> None:
         root_w, root_h, max_w, max_h = packet[1:5]
         log("server has resized the desktop to: %sx%s (max %sx%s)", root_w, root_h, max_w, max_h)
         self.server_max_desktop_size = max_w, max_h
@@ -344,7 +349,7 @@ class DisplayClient(StubClientMixin):
             self.may_adjust_scaling()
 
 
-    def may_adjust_scaling(self):
+    def may_adjust_scaling(self) -> None:
         log("may_adjust_scaling() server_is_desktop=%s, desktop_fullscreen=%s",
             self.server_is_desktop, self.desktop_fullscreen)
         if self.server_is_desktop and not self.desktop_fullscreen:
@@ -387,14 +392,14 @@ class DisplayClient(StubClientMixin):
         #to use the same scale for both axes:
         #self.xscale = mint(max(x, y))
         #self.yscale = self.xscale
-        summary = "Desktop scaling adjusted to accomodate the server"
+        summary = "Desktop scaling adjusted to accommodate the server"
         xstr = ("%.3f" % self.xscale).rstrip("0")
         ystr = ("%.3f" % self.yscale).rstrip("0")
         messages = [
             "server desktop size is %ix%i" % (max_w, max_h),
             "using scaling factor %s x %s" % (xstr, ystr),
             ]
-        self.may_notify(XPRA_SCALING_NOTIFICATION_ID, summary, "\n".join(messages), icon_name="scaling")
+        self.may_notify(NotificationID.SCALING, summary, "\n".join(messages), icon_name="scaling")
         scalinglog.warn("Warning: %s", summary)
         for m in messages:
             scalinglog.warn(" %s", m)
@@ -403,10 +408,10 @@ class DisplayClient(StubClientMixin):
 
     ######################################################################
     # screen scaling:
-    def fsx(self, v):
+    def fsx(self, v) -> float:
         """ convert X coordinate from server to client """
         return v*self.xscale
-    def fsy(self, v):
+    def fsy(self, v) -> float:
         """ convert Y coordinate from server to client """
         return v*self.yscale
     def sx(self, v) -> int:
@@ -415,10 +420,10 @@ class DisplayClient(StubClientMixin):
     def sy(self, v) -> int:
         """ convert Y coordinate from server to client """
         return round(self.fsy(v))
-    def srect(self, x, y, w, h):
+    def srect(self, x, y, w, h) -> Tuple[int, int, int, int]:
         """ convert rectangle coordinates from server to client """
         return self.sx(x), self.sy(y), self.sx(w), self.sy(h)
-    def sp(self, x, y):
+    def sp(self, x, y) -> Tuple[float, float]:
         """ convert X,Y coordinates from server to client """
         return self.sx(x), self.sy(y)
 
@@ -428,31 +433,31 @@ class DisplayClient(StubClientMixin):
     def cy(self, v) -> int:
         """ convert Y coordinate from client to server """
         return round(v/self.yscale)
-    def crect(self, x, y, w, h):
+    def crect(self, x, y, w, h) -> Tuple[int, int, int, int]:
         """ convert rectangle coordinates from client to server """
         return self.cx(x), self.cy(y), self.cx(w), self.cy(h)
-    def cp(self, x, y):
+    def cp(self, x, y) -> Tuple[int, int]:
         """ convert X,Y coordinates from client to server """
         return self.cx(x), self.cy(y)
 
 
     ######################################################################
     # desktop, screen and scaling:
-    def desktops_changed(self, *args):
+    def desktops_changed(self, *args) -> None:
         workspacelog("desktops_changed%s", args)
         self.screen_size_changed(*args)
 
-    def workspace_changed(self, *args):
+    def workspace_changed(self, *args) -> None:
         workspacelog("workspace_changed%s", args)
         for win in self._id_to_window.values():
             win.workspace_changed()
 
-    def screen_size_changed(self, *args):
+    def screen_size_changed(self, *args) -> None:
         log("screen_size_changed(%s) timer=%s", args, self.screen_size_change_timer)
         if self.screen_size_change_timer:
             return
         #update via timer so the data is more likely to be final (up to date) when we query it,
-        #some properties (like _NET_WORKAREA for X11 clients via xposix "ClientExtras") may
+        #some properties (like _NET_WORKAREA for X11 clients via posix gui "ClientExtras") may
         #trigger multiple calls to screen_size_changed, delayed by some amount
         #(sometimes up to 1s..)
         delay = 1000
@@ -463,7 +468,7 @@ class DisplayClient(StubClientMixin):
         self.screen_size_change_timer = self.timeout_add(delay, self.do_process_screen_size_change)
 
     def do_process_screen_size_change(self):
-        self.screen_size_change_timer = None
+        self.screen_size_change_timer = 0
         self.update_screen_size()
         log("do_process_screen_size_change() MONITOR_CHANGE_REINIT=%s, REINIT_WINDOWS=%s",
             MONITOR_CHANGE_REINIT, REINIT_WINDOWS)
@@ -475,7 +480,7 @@ class DisplayClient(StubClientMixin):
             self.reinit_window_icons()
 
 
-    def get_screen_settings(self):
+    def get_screen_settings(self) -> Tuple:
         u_root_w, u_root_h = self.get_root_size()
         root_w, root_h = self.cp(u_root_w, u_root_h)
         self._current_screen_sizes = self.get_screen_sizes()
@@ -503,8 +508,8 @@ class DisplayClient(StubClientMixin):
         monitors = self.get_monitors_info()
         return (root_w, root_h, sss, ndesktops, desktop_names, u_root_w, u_root_h, xdpi, ydpi, rrate, monitors)
 
-    def update_screen_size(self):
-        self.screen_size_change_timer = None
+    def update_screen_size(self) -> None:
+        self.screen_size_change_timer = 0
         screen_settings = self.get_screen_settings()
         log("update_screen_size()     new settings=%s", screen_settings)
         log("update_screen_size() current settings=%s", self._last_screen_settings)
@@ -514,7 +519,36 @@ class DisplayClient(StubClientMixin):
         root_w, root_h, sss = screen_settings[:3]
         log.info("sending updated screen size to server: %sx%s", root_w, root_h)
         log_screen_sizes(root_w, root_h, sss)
-        if self.server_desktop_size:
+        if "configure-display" in self.server_packet_types:
+            #new packet format:
+            root_w, root_h = screen_settings[:2]
+            ndesktops, desktop_names = screen_settings[3:5]
+            u_root_w, u_root_h = screen_settings[5:7]
+            xdpi, ydpi = screen_settings[7:9]
+            rrate, monitors = screen_settings[9:11]
+            attrs = {
+                "desktop-size"          : (root_w, root_h),
+                "desktop-size-unscaled" : (u_root_w, u_root_h),
+                "monitors"              : monitors,
+                "dpi"                   : {"x" : xdpi, "y" : ydpi},
+                }
+            icc = self.get_icc_info()
+            dicc = self.get_display_icc_info()
+            if icc or dicc:
+                attrs["icc"] = {
+                    "global"    : icc,
+                    "display"   : dicc,
+                    }
+            if ndesktops:
+                attrs["desktops"] = ndesktops
+                attrs["desktop-names"] = desktop_names or ()
+            if rrate:
+                attrs["vrefresh"] = rrate
+            log(f"configure-display: {attrs}")
+            self.send("configure-display", attrs)
+        elif self.server_desktop_size:
+            log(f"sending legacy desktop-size packet: {screen_settings}")
+            #legacy packet:
             self.send("desktop_size", *screen_settings)
         self._last_screen_settings = screen_settings
         #update the max packet size (may have gone up):
@@ -527,37 +561,37 @@ class DisplayClient(StubClientMixin):
         return get_ydpi()
 
 
-    def scaleup(self):
+    def scaleup(self) -> None:
         scaling = max(self.xscale, self.yscale)
         options = scaleup_value(scaling)
         scalinglog("scaleup() options=%s", options)
         if options:
             self._scaleto(min(options))
 
-    def scaledown(self):
+    def scaledown(self) -> None:
         scaling = min(self.xscale, self.yscale)
         options = scaledown_value(scaling)
         scalinglog("scaledown() options=%s", options)
         if options:
             self._scaleto(max(options))
 
-    def _scaleto(self, new_scaling):
+    def _scaleto(self, new_scaling) -> None:
         scaling = max(self.xscale, self.yscale)
         scalinglog("_scaleto(%s) current value=%s", r4cmp(new_scaling, 1000)/1000.0, r4cmp(scaling, 1000)/1000.0)
         if new_scaling>0:
             self.scale_change(new_scaling/scaling, new_scaling/scaling)
 
-    def scalingoff(self):
+    def scalingoff(self) -> None:
         self.scaleset(1, 1)
 
-    def scalereset(self):
+    def scalereset(self) -> None:
         self.scaleset(*self.initial_scaling)
 
-    def scaleset(self, xscale=1, yscale=1):
+    def scaleset(self, xscale=1.0, yscale=1.0) -> None:
         scalinglog("scaleset(%s, %s) current scaling: %s, %s", xscale, yscale, self.xscale, self.yscale)
         self.scale_change(xscale/self.xscale, yscale/self.yscale)
 
-    def scale_change(self, xchange=1, ychange=1):
+    def scale_change(self, xchange=1.0, ychange=1.0) -> None:
         scalinglog("scale_change(%s, %s)", xchange, ychange)
         if self.server_is_desktop and self.desktop_fullscreen:
             scalinglog("scale_change(%s, %s) ignored, fullscreen shadow mode is active", xchange, ychange)
@@ -599,7 +633,7 @@ class DisplayClient(StubClientMixin):
                 "the scaled client screen %i x %i -> %i x %i" % (root_w, root_h, sw, sh),
                 " would overflow the server's screen: %i x %i" % (maxw, maxh),
                 ]
-            self.may_notify(XPRA_SCALING_NOTIFICATION_ID, summary, "\n".join(messages), icon_name="scaling")
+            self.may_notify(NotificationID.SCALING, summary, "\n".join(messages), icon_name="scaling")
             scalinglog.warn("Warning: %s", summary)
             for m in messages:
                 scalinglog.warn(" %s", m)
@@ -609,7 +643,7 @@ class DisplayClient(StubClientMixin):
         scalinglog("scale_change new scaling: %sx%s, change: %sx%s", self.xscale, self.yscale, xchange, ychange)
         self.scale_reinit(xchange, ychange)
 
-    def scale_reinit(self, xchange=1.0, ychange=1.0):
+    def scale_reinit(self, xchange=1.0, ychange=1.0) -> None:
         #wait at least one second before changing again:
         self.scale_change_embargo = monotonic()+SCALING_EMBARGO_TIME
         if fequ(self.xscale, self.yscale):
@@ -633,6 +667,6 @@ class DisplayClient(StubClientMixin):
         self.emit("scaling-changed")
 
 
-    def init_authenticated_packet_handlers(self):
+    def init_authenticated_packet_handlers(self) -> None:
         self.add_packet_handler("show-desktop", self._process_show_desktop)
         self.add_packet_handler("desktop_size", self._process_desktop_size)
