@@ -1,11 +1,13 @@
 # This file is part of Xpra.
-# Copyright (C) 2016-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2016-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from xpra.util import envbool
+from typing import Callable, Tuple
+
+from xpra.util import envbool, is_valid_hostname
 from xpra.net.websockets.common import make_websocket_accept_hash
-from xpra.net.http_handler import HTTPRequestHandler
+from xpra.net.http.http_handler import HTTPRequestHandler, AUTH_USERNAME, AUTH_PASSWORD
 from xpra.log import Logger
 
 log = Logger("network", "websocket")
@@ -15,7 +17,7 @@ WEBSOCKET_ONLY_UPGRADE = envbool("XPRA_WEBSOCKET_ONLY_UPGRADE", False)
 # HyBi-07 report version 7
 # HyBi-08 - HyBi-12 report version 8
 # HyBi-13 reports version 13
-SUPPORT_HyBi_PROTOCOLS = ("7", "8", "13")
+SUPPORT_HyBi_PROTOCOLS : Tuple[str,...] = ("7", "8", "13")
 
 
 class WebSocketRequestHandler(HTTPRequestHandler):
@@ -26,45 +28,60 @@ class WebSocketRequestHandler(HTTPRequestHandler):
                  web_root="/usr/share/xpra/www/",
                  http_headers_dir="/etc/xpra/http-headers",
                  script_paths=None,
-                 redirect_https=False):
-        self.new_websocket_client = new_websocket_client
+                 redirect_https=False,
+                 username=AUTH_USERNAME, password=AUTH_PASSWORD,
+                 ):
+        self.new_websocket_client : Callable = new_websocket_client
         self.only_upgrade = WEBSOCKET_ONLY_UPGRADE
         self.redirect_https = redirect_https
-        super().__init__(sock, addr, web_root, http_headers_dir, script_paths)
+        super().__init__(sock, addr,
+                         web_root, http_headers_dir, script_paths,
+                         username, password)
 
-    def handle_websocket(self):
+    def handle_websocket(self) -> None:
         log("handle_websocket() calling %s, request=%s (%s)",
             self.new_websocket_client, self.request, type(self.request))
         log("headers:")
         for k,v in self.headers.items():
-            log(" %s=%s", k, v)
-        ver = self.headers.get('Sec-WebSocket-Version')
-        if ver is None:
-            raise Exception("Missing Sec-WebSocket-Version header")
+            log(f" {k}={v}")
+        ver = self.headers.get("Sec-WebSocket-Version", "")
+        if not ver:
+            raise ValueError("Missing Sec-WebSocket-Version header")
 
         if ver not in SUPPORT_HyBi_PROTOCOLS:
-            raise Exception("Unsupported protocol version %s" % ver)
+            raise ValueError(f"Unsupported protocol version {ver}")
 
         protocols = self.headers.get("Sec-WebSocket-Protocol", "").split(",")
         if "binary" not in protocols:
-            raise Exception("client does not support 'binary' protocol")
+            raise ValueError("client does not support 'binary' protocol")
 
-        key = self.headers.get("Sec-WebSocket-Key")
-        if key is None:
-            raise Exception("Missing Sec-WebSocket-Key header")
-        for upgrade_string in (
+        key = self.headers.get("Sec-WebSocket-Key", "")
+        if not key:
+            raise ValueError("Missing Sec-WebSocket-Key header")
+        accept = make_websocket_accept_hash(key)
+        log(f"websocket hash for key {key!r} = {accept!r}")
+        self.write_byte_strings(
             b"HTTP/1.1 101 Switching Protocols",
             b"Upgrade: websocket",
             b"Connection: Upgrade",
-            b"Sec-WebSocket-Accept: %s" % make_websocket_accept_hash(key),
+            b"Sec-WebSocket-Accept: %s" % accept,
             b"Sec-WebSocket-Protocol: %s" % b"binary",
             b"",
-            ):
-            self.wfile.write(b"%s\r\n" % upgrade_string)
-        self.wfile.flush()
+            b"",
+            )
         self.new_websocket_client(self)
+        #don't use our finish method that closes the socket,
+        #but do call the superclass's finish() method:
+        self.finish = super().finish
 
-    def do_GET(self):
+    def write_byte_strings(self, *bstrings):
+        bdata = b"\r\n".join(bstrings)
+        self.wfile.write(bdata)
+        self.wfile.flush()
+
+
+    def do_GET(self) -> None:
+        log(f"do_GET() path={self.path!r}, headers={self.headers!r}")
         upgrade_requested = (self.headers.get('upgrade') or "").lower() == 'websocket'
         if self.only_upgrade or upgrade_requested:
             if not upgrade_requested:
@@ -75,15 +92,15 @@ class WebSocketRequestHandler(HTTPRequestHandler):
             except Exception as e:
                 log("do_GET()", exc_info=True)
                 log.error("Error: cannot handle websocket upgrade:")
-                log.error(" %s", e)
-                self.send_error(403, "failed to handle websocket: %s" % e)
+                log.estr(e)
+                self.send_error(403, f"failed to handle websocket: {e}")
             return
         if self.headers.get("Upgrade-Insecure-Requests", "")=="1" and self.redirect_https:
             self.do_redirect_https()
             return
         super().do_GET()
 
-    def do_HEAD(self):
+    def do_HEAD(self) -> None:
         if self.only_upgrade:
             self.send_error(405, "Method Not Allowed")
             return
@@ -92,22 +109,38 @@ class WebSocketRequestHandler(HTTPRequestHandler):
             return
         super().do_HEAD()
 
-    def do_redirect_https(self):
-        if not self.headers["Host"]:
-            self.send_error(400, "Client did not send Host: header")
-            return
+    def do_redirect_https(self) -> None:
         server_address = self.headers["Host"]
-        for upgrade_string in (
+        if not server_address:
+            log.warn("Warning: cannot redirect to https without a 'Host' header")
+            self.send_error(400, "Client did not send a 'Host' header")
+            return
+        parts = server_address.split(":")
+        if len(parts) == 2:
+            host = parts[0]
+        else:
+            host = server_address
+        if not is_valid_hostname(host):
+            log.warn("Warning: cannot redirect to https using an invalid hostname")
+            log.warn(f" {host!r}")
+            self.send_error(400, "Client specified an invalid 'Host' header")
+            return
+        self.write_byte_strings(
             b"HTTP/1.1 301 Moved Permanently",
             b"Connection: close",
             b"Location: https://%s%s" % (bytes(server_address, "utf-8"), bytes(self.path, "utf-8")),
             b"",
-            ):
-            self.wfile.write(b"%s\r\n" % upgrade_string)
-        self.wfile.flush()
+            b"",
+            )
 
-    def handle_request(self):
+    def handle_request(self) -> None:
         if self.only_upgrade:
             self.send_error(405, "Method Not Allowed")
             return
         super().handle_request()
+
+    def finish(self) -> None:
+        super().finish()
+        log(f"finish() close_connection={self.close_connection}, connection={self.connection}")
+        if self.close_connection:
+            self.connection.close()
