@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2013-2022 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2013-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -17,36 +17,41 @@ import binascii
 import threading
 from time import monotonic, sleep
 from subprocess import PIPE, Popen
+from typing import Optional, Tuple, Dict, Any, Set, List, Callable
+from threading import Thread
 
-# only mininal imports go at the top
+# only minimal imports go at the top
 # so that this file can be included everywhere
-# without too many side-effects
+# without too many side effects
 # pylint: disable=import-outside-toplevel
 
-SIGNAMES = {}
+SIGNAMES : Dict[int,str] = {}
 for signame in (sig for sig in dir(signal) if sig.startswith("SIG") and not sig.startswith("SIG_")):
-    SIGNAMES[getattr(signal, signame)] = signame
+    try:
+        SIGNAMES[int(getattr(signal, signame))] = signame
+    except ValueError:
+        pass
 
 
-WIN32 = sys.platform.startswith("win")
-OSX = sys.platform.startswith("darwin")
-LINUX = sys.platform.startswith("linux")
-NETBSD = sys.platform.startswith("netbsd")
-OPENBSD = sys.platform.startswith("openbsd")
-FREEBSD  = sys.platform.startswith("freebsd")
+WIN32 : bool = sys.platform.startswith("win")
+OSX : bool = sys.platform.startswith("darwin")
+LINUX : bool = sys.platform.startswith("linux")
+NETBSD : bool = sys.platform.startswith("netbsd")
+OPENBSD : bool = sys.platform.startswith("openbsd")
+FREEBSD : bool = sys.platform.startswith("freebsd")
 
-POSIX = os.name=="posix"
+POSIX : bool = os.name=="posix"
 
-BITS = struct.calcsize(b"P")*8
+BITS : int = struct.calcsize(b"P")*8
 
 
 main_thread = threading.current_thread()
-def is_main_thread():
+def is_main_thread() -> bool:
     return threading.current_thread()==main_thread
 
 
-def get_frame_info(ignore_threads=()):
-    info = {
+def get_frame_info(ignore_threads:Tuple[Thread,...]=()) -> Dict[Any,Any]:
+    info : Dict[Any,Any] = {
         "count"        : threading.active_count() - len(ignore_threads),
         }
     try:
@@ -55,15 +60,15 @@ def get_frame_info(ignore_threads=()):
             if x is None:
                 return ""
             return str(x)
-        thread_ident = {}
+        thread_ident : Dict[Optional[int],Optional[str]] = {}
         for t in threading.enumerate():
             if t not in ignore_threads:
-                thread_ident[t.ident] = t.getName()
+                thread_ident[t.ident] = t.name
             else:
                 thread_ident[t.ident] = None
         thread_ident.update({
-                threading.current_thread().ident    : "info",
-                main_thread.ident                   : "main",
+                threading.current_thread().ident  : "info",
+                main_thread.ident                 : "main",
                 })
         frames = sys._current_frames()  #pylint: disable=protected-access
         stack = None
@@ -85,7 +90,7 @@ def get_frame_info(ignore_threads=()):
         get_util_logger().error("failed to get frame info: %s", e)
     return info
 
-def get_info_env():
+def get_info_env() -> Dict[str,str]:
     filtered_env = os.environ.copy()
     if filtered_env.get('XPRA_PASSWORD'):
         filtered_env['XPRA_PASSWORD'] = "*****"
@@ -93,9 +98,9 @@ def get_info_env():
         filtered_env['XPRA_ENCRYPTION_KEY'] = "*****"
     return filtered_env
 
-def get_sysconfig_info():
+def get_sysconfig_info() -> Dict[str,Any]:
     import sysconfig
-    sysinfo = {}
+    sysinfo : Dict[str,Any] = {}
     log = get_util_logger()
     for attr in (
         "platform",
@@ -120,7 +125,10 @@ def get_sysconfig_info():
 def strtobytes(x) -> bytes:
     if isinstance(x, bytes):
         return x
-    return str(x).encode("latin1")
+    try:
+        return str(x).encode("latin1")
+    except UnicodeEncodeError:
+        return str(x).encode("utf8")
 def bytestostr(x) -> str:
     if isinstance(x, (bytes, bytearray)):
         return x.decode("latin1")
@@ -147,7 +155,7 @@ def memoryview_to_bytes(v) -> bytes:
     return strtobytes(v)
 
 
-def set_proc_title(title):
+def set_proc_title(title) -> None:
     try:
         import setproctitle  #pylint: disable=import-outside-toplevel
         setproctitle.setproctitle(title)  #@UndefinedVariable pylint: disable=c-extension-no-member
@@ -192,7 +200,7 @@ def get_home_for_uid(uid) -> str:
             pass
     return ""
 
-def get_groups(username):
+def get_groups(username) -> List[str]:
     if POSIX:
         import grp      #@UnresolvedImport
         return [gr.gr_name for gr in grp.getgrall() if username in gr.gr_mem]
@@ -226,13 +234,14 @@ def platform_release(release):
 def platform_name(sys_platform=sys.platform, release=None) -> str:
     if not sys_platform:
         return "unknown"
-    PLATFORMS = {"win32"    : "Microsoft Windows",
-                 "cygwin"   : "Windows/Cygwin",
-                 "linux.*"  : "Linux",
-                 "darwin"   : "Mac OS X",
-                 "freebsd.*": "FreeBSD",
-                 "os2"      : "OS/2",
-                 }
+    PLATFORMS = {
+        "win32"    : "Microsoft Windows",
+        "cygwin"   : "Windows/Cygwin",
+        "linux.*"  : "Linux",
+        "darwin"   : "Mac OS X",
+        "freebsd.*": "FreeBSD",
+        "os2"      : "OS/2",
+        }
     def rel(v):
         values = [v]
         if isinstance(release, (tuple, list)):
@@ -247,10 +256,6 @@ def platform_name(sys_platform=sys.platform, release=None) -> str:
     return rel(sys_platform)
 
 
-def get_rand_chars(l=16, chars=b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ") -> bytes:
-    import random
-    return b"".join(chars[random.randint(0, len(chars)-1):][:1] for _ in range(l))
-
 def get_hex_uuid() -> str:
     return uuid.uuid4().hex
 
@@ -266,16 +271,17 @@ def get_machine_id() -> str:
     v = ""
     if POSIX:
         for filename in ("/etc/machine-id", "/var/lib/dbus/machine-id"):
-            v = load_binary_file(filename)
-            if v is not None:
+            b = load_binary_file(filename)
+            if b is not None:
+                v = bytestostr(b)
                 break
     elif WIN32:
-        v = uuid.getnode()
-    return bytestostr(v).strip("\n\r")
+        v = str(uuid.getnode())
+    return v.strip("\n\r")
 
 def get_user_uuid() -> str:
     """
-        Try to generate a uuid string which is unique to this user.
+        Try to generate an uuid string which is unique to this user.
         (relies on get_machine_id to uniquely identify a machine)
     """
     user_uuid = os.environ.get("XPRA_USER_UUID")
@@ -298,6 +304,8 @@ def get_user_uuid() -> str:
 def is_X11() -> bool:
     if OSX or WIN32:
         return False
+    if os.environ.get("XPRA_NOX11", "")=="1":
+        return False
     try:
         from xpra.x11.gtk3.gdk_bindings import is_X11_Display   #@UnresolvedImport
         return is_X11_Display()
@@ -323,7 +331,7 @@ def restore_script_env(env):
 
 
 _saved_env = os.environ.copy()
-def get_saved_env():
+def get_saved_env() -> Dict[str,str]:
     return _saved_env.copy()
 
 def get_saved_env_var(var, default=None):
@@ -332,7 +340,7 @@ def get_saved_env_var(var, default=None):
 def is_Wayland() -> bool:
     return _is_Wayland(_saved_env)
 
-def _is_Wayland(env : dict) -> bool:
+def _is_Wayland(env : Dict) -> bool:
     backend = env.get("GDK_BACKEND", "")
     if backend=="wayland":
         return True
@@ -351,9 +359,9 @@ def is_distribution_variant(variant=b"Debian") -> bool:
         pass
     try:
         d = get_linux_distribution()[0]
-        if d==variant:
+        if d==bytestostr(variant):
             return True
-        if variant==b"RedHat" and d.startswith(variant):
+        if variant==b"RedHat" and d.startswith(bytestostr(variant)):
             return True
     except Exception:
         pass
@@ -372,14 +380,14 @@ def get_distribution_version_id() -> str:
         pass
     return ""
 
-os_release_file_data = False
+os_release_file_data : Optional[bytes] = None
 def load_os_release_file() -> bytes:
     global os_release_file_data
-    if os_release_file_data is False:
+    if os_release_file_data is None:
         try:
-            os_release_file_data = load_binary_file("/etc/os-release")
+            os_release_file_data = load_binary_file("/etc/os-release") or b""
         except OSError: # pragma: no cover
-            os_release_file_data = None
+            os_release_file_data = b""
     return os_release_file_data
 
 def is_Ubuntu() -> bool:
@@ -421,12 +429,13 @@ def is_arm() -> bool:
     return platform.uname()[4].startswith("arm")
 
 
-_linux_distribution = None
-def get_linux_distribution():
+_linux_distribution = ("", "", "")
+def get_linux_distribution() -> Tuple[str,str,str]:
     global _linux_distribution
-    if LINUX and not _linux_distribution:
-        #linux_distribution is deprecated in Python 3.5 and it causes warnings,
-        #so use our own code first:
+    if LINUX and _linux_distribution==("", "", ""):
+        # linux_distribution is deprecated in Python 3.5,
+        # and it causes warnings,
+        # so we use our own code first:
         cmd = ["lsb_release", "-a"]
         try:
             p = Popen(cmd, stdout=PIPE, stderr=PIPE)
@@ -440,14 +449,15 @@ def get_linux_distribution():
                 _linux_distribution = ("unknown", "unknown", "unknown")
         else:
             d = {}
-            for line in strtobytes(out).splitlines():
-                line = bytestostr(line)
+            for line in bytestostr(out).splitlines():
                 parts = line.rstrip("\n\r").split(":", 1)
                 if len(parts)==2:
                     d[parts[0].lower().replace(" ", "_")] = parts[1].strip()
-            v = [d.get(x) for x in ("distributor_id", "release", "codename")]
-            if None not in v:
-                return tuple([bytestostr(x) for x in v])
+            _linux_distribution = (
+                d.get("distributor_id","unknown"),
+                d.get("release", "unknown"),
+                d.get("codename", "unknown"),
+            )
     return _linux_distribution
 
 def is_unity() -> bool:
@@ -499,16 +509,16 @@ def do_get_generic_os_name() -> str:
     return sys.platform     # pragma: no cover
 
 
-def filedata_nocrlf(filename) -> str:
+def filedata_nocrlf(filename:str) -> bytes:
     v = load_binary_file(filename)
     if v is None:
         get_util_logger().error(f"failed to load {filename!r}")
-        return None
+        return b""
     return v.strip(b"\n\r")
 
 def load_binary_file(filename) -> bytes:
     if not os.path.exists(filename):
-        return None
+        return b""
     try:
         with open(filename, "rb") as f:
             return f.read()
@@ -516,19 +526,23 @@ def load_binary_file(filename) -> bytes:
         get_util_logger().debug(f"load_binary_file({filename})", exc_info=True)
         get_util_logger().warn(f"Warning: failed to load {filename!r}")
         get_util_logger().warn(f" {e}")
-        return None
+        return b""
 
-def get_proc_cmdline(pid):
+def get_proc_cmdline(pid:int) -> Tuple[str, ...]:
     if pid and LINUX:
         #try to find the command via /proc:
         proc_cmd_line = os.path.join("/proc", f"{pid}", "cmdline")
         if os.path.exists(proc_cmd_line):
-            return load_binary_file(proc_cmd_line).rstrip(b"\0").split(b"\0")
-    return None
+            cmdline = load_binary_file(proc_cmd_line).rstrip(b"\0").split(b"\0")
+            try:
+                return tuple(x.decode() for x in cmdline)
+            except Exception:
+                return tuple(bytestostr(x) for x in cmdline)
+    return ()
 
-def parse_encoded_bin_data(data):
+def parse_encoded_bin_data(data:str) -> bytes:
     if not data:
-        return None
+        return b""
     header = bytestostr(data).lower()[:10]
     if header.startswith("0x"):
         return binascii.unhexlify(data[2:])
@@ -544,18 +558,18 @@ def parse_encoded_bin_data(data):
             return base64.b64decode(data)
         except Exception:
             pass
-    return None
+    return b""
 
 
 #here so we can override it when needed
-def force_quit(status=1):
+def force_quit(status=1) -> None:
     os._exit(status)  #pylint: disable=protected-access
 
 
 def no_idle(fn, *args, **kwargs):
     fn(*args, **kwargs)
-def register_SIGUSR_signals(idle_add=no_idle):
-    if not os.name=="posix":
+def register_SIGUSR_signals(idle_add=no_idle) -> None:
+    if os.name!="posix":
         return
     from xpra.util import dump_all_frames, dump_gc_frames
     def sigusr1(*_args):
@@ -570,7 +584,7 @@ def register_SIGUSR_signals(idle_add=no_idle):
     signal.signal(signal.SIGUSR2, sigusr2)
 
 
-def livefds():
+def livefds() -> Set[int]:
     live = set()
     try:
         MAXFD = os.sysconf("SC_OPEN_MAX")
@@ -587,34 +601,36 @@ def livefds():
     return live
 
 
-def use_tty():
+def use_tty() -> bool:
     from xpra.util import envbool
     if envbool("XPRA_NOTTY", False):
         return False
     from xpra.platform.gui import use_stdin
     return use_stdin()
 
-def use_gui_prompt():
+def use_gui_prompt() -> bool:
     return WIN32 or OSX or not use_tty()
 
 
-def shellsub(s, subs=None):
+def shellsub(s : str, subs=None) -> str:
     """ shell style string substitution using the dictionary given """
     if subs:
         for var,value in subs.items():
             try:
                 if isinstance(s, bytes):
-                    s = s.replace(f"${var}".encode(), str(value).encode())
-                    s = s.replace(("${%s}" % var).encode(), str(value).encode())
+                    vbin = str(value).encode()
+                    s = s.replace(f"${var}".encode(), vbin)
+                    s = s.replace(("${%s}" % var).encode(), vbin)
                 else:
-                    s = s.replace(f"${var}", str(value))
-                    s = s.replace("${%s}" % var, str(value))
+                    vstr = str(value)
+                    s = s.replace(f"${var}", vstr)
+                    s = s.replace("${%s}" % var, vstr)
             except (TypeError, ValueError):
-                raise Exception(f"failed to substitute {var!r} with value {value!r} ({type(value)}) in {s!r}") from None
+                raise ValueError(f"failed to substitute {var!r} with value {value!r} ({type(value)}) in {s!r}") from None
     return s
 
 
-def osexpand(s, actual_username="", uid=0, gid=0, subs=None):
+def osexpand(s : str, actual_username="", uid=0, gid=0, subs=None) -> str:
     if not s:
         return s
     def expanduser(s):
@@ -633,7 +649,7 @@ def osexpand(s, actual_username="", uid=0, gid=0, subs=None):
             "GID"   : gid or os.getegid(),
             })
         if not OSX:
-            from xpra.platform.xposix.paths import get_runtime_dir
+            from xpra.platform.posix.paths import get_runtime_dir
             rd = get_runtime_dir()
             if rd and "XDG_RUNTIME_DIR" not in os.environ:
                 d["XDG_RUNTIME_DIR"] = rd
@@ -648,9 +664,9 @@ def osexpand(s, actual_username="", uid=0, gid=0, subs=None):
     return os.path.expandvars(expanduser(shellsub(expanduser(s), ssub)))
 
 
-def path_permission_info(filename, ftype=None):
+def path_permission_info(filename : str, ftype=None) -> Tuple[str, ...]:
     if not POSIX:
-        return []
+        return ()
     info = []
     try:
         stat_info = os.stat(filename)
@@ -668,16 +684,17 @@ def path_permission_info(filename, ftype=None):
         info.append(f"ownership {user}:{group}")
     except Exception as e:
         info.append(f"failed to query path information for {filename!r}: {e}")
-    return info
+    return tuple(info)
 
 
-#code to temporarily redirect stderr and restore it afterwards, adapted from:
-#http://stackoverflow.com/questions/5081657/how-do-i-prevent-a-c-shared-library-to-print-on-stdout-in-python
-#used by the sound code to get rid of the stupid gst warning below:
-#"** Message: pygobject_register_sinkfunc is deprecated (GstObject)"
-#ideally we would redirect to a buffer so we could still capture and show these messages in debug out
+# code to temporarily redirect stderr and restore it afterwards, adapted from:
+# http://stackoverflow.com/questions/5081657/how-do-i-prevent-a-c-shared-library-to-print-on-stdout-in-python
+# used by the audio code to get rid of the stupid gst warning below:
+# "** Message: pygobject_register_sinkfunc is deprecated (GstObject)"
+# ideally we would redirect to a buffer,
+# so we could still capture and show these messages in debug out
 class HideStdErr:
-    __slots__ = ("savedstderr")
+    __slots__ = ("savedstderr", )
     def __init__(self, *_args):
         self.savedstderr = None
 
@@ -690,14 +707,20 @@ class HideStdErr:
         devnull = os.open(os.devnull, os.O_WRONLY)
         os.dup2(devnull, 2)
         os.close(devnull)
-        sys.stderr = os.fdopen(self.savedstderr, 'w')
+        try:
+            sys.stderr = os.fdopen(self.savedstderr, 'w')
+        except OSError as e:
+            try:
+                sys.stderr.write(f"failed to replace stderr: {e}\n")
+            except:
+                pass
 
     def __exit__(self, *_args):
         if self.savedstderr is not None:
             os.dup2(self.savedstderr, 2)
 
 class HideSysArgv:
-    __slots__ = ("savedsysargv")
+    __slots__ = ("savedsysargv", )
     def __init__(self, *_args):
         self.savedsysargv = None
 
@@ -711,11 +734,13 @@ class HideSysArgv:
 
 
 class OSEnvContext:
-    __slots__ = ("env")
-    def __init__(self):
+    __slots__ = ("env", "kwargs")
+    def __init__(self, **kwargs):
         self.env = {}
+        self.kwargs = kwargs
     def __enter__(self):
         self.env = os.environ.copy()
+        os.environ.update(self.kwargs)
     def __exit__(self, *_args):
         os.environ.clear()
         os.environ.update(self.env)
@@ -740,7 +765,8 @@ class nomodule_context:
         self.module_name = module_name
     def __enter__(self):
         self.saved_module = sys.modules.get(self.module_name)
-        sys.modules[self.module_name] = None
+        # noinspection PyTypeChecker
+        sys.modules[self.module_name] = None        # type: ignore[assignment]
     def __exit__(self, *_args):
         if sys.modules.get(self.module_name) is None:
             if self.saved_module is None:
@@ -770,7 +796,7 @@ def disable_stdout_buffering():
     gc.garbage.append(sys.stdout)
     sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
 
-def setbinarymode(fd):
+def setbinarymode(fd:int):
     if WIN32:
         #turn on binary mode:
         try:
@@ -779,14 +805,11 @@ def setbinarymode(fd):
         except OSError:
             get_util_logger().error("setting stdin to binary mode failed", exc_info=True)
 
-def find_lib_ldconfig(libname):
+def find_lib_ldconfig(libname:str) -> str:
     libname = re.escape(libname)
-
     arch_map = {"x86_64": "libc6,x86-64"}
     arch = arch_map.get(os.uname()[4], "libc6")
-
     pattern = r'^\s+lib%s\.[^\s]+ \(%s(?:,.*?)?\) => (.*lib%s[^\s]+)' % (libname, arch, libname)
-
     #try to find ldconfig first, which may not be on the $PATH
     #(it isn't on Debian..)
     ldconfig = "ldconfig"
@@ -798,16 +821,15 @@ def find_lib_ldconfig(libname):
     import subprocess
     p = subprocess.Popen(f"{ldconfig} -p", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
     data = bytestostr(p.communicate()[0])
-
     libpath = re.search(pattern, data, re.MULTILINE)        #@UndefinedVariable
     if libpath:
-        libpath = libpath.group(1)
-    return libpath
+        return libpath.group(1)
+    return ""
 
-def find_lib(libname):
-    #it would be better to rely on dlopen to find the paths
-    #but I cannot find a way of getting ctypes to tell us the path
-    #it found the library in
+def find_lib(libname:str):
+    # it would be better to rely on dlopen to find the paths,
+    # but I cannot find a way of getting ctypes to tell us the path
+    # it found the library in
     assert POSIX
     libpaths = os.environ.get("LD_LIBRARY_PATH", "").split(":")
     if BITS==64 and os.path.exists("/usr/lib64"):
@@ -823,7 +845,7 @@ def find_lib(libname):
     return None
 
 
-def pollwait(process, timeout=5):
+def pollwait(process, timeout=5) -> Optional[int]:
     start = monotonic()
     v = None
     while monotonic()-start<timeout:
@@ -834,7 +856,7 @@ def pollwait(process, timeout=5):
     return v
 
 
-def find_in_PATH(command):
+def find_in_PATH(command) -> Optional[str]:
     path = os.environ.get("PATH", None)
     if not path:
         return None
@@ -845,21 +867,29 @@ def find_in_PATH(command):
             return f
     return None
 
-def which(command):
+
+def get_which_impl() -> Callable[[Any],Optional[str]]:
     try:
-        from shutil import which as find_executable
+        from shutil import which
+        return which
     except ImportError:
-        try:
-            from distutils.spawn import find_executable  # pylint: disable=deprecated-module
-        except ImportError:
-            find_executable = find_in_PATH
+        pass
+    try:
+        from distutils.spawn import find_executable # pylint: disable=deprecated-module
+        return find_executable
+    except ImportError:
+        return find_in_PATH
+
+
+def which(command) -> Optional[str]:
+    find_executable = get_which_impl()
     try:
         return find_executable(command)
     except Exception:
         get_util_logger().debug(f"find_executable({command})", exc_info=True)
         return None
 
-def get_status_output(*args, **kwargs):
+def get_status_output(*args, **kwargs) -> Tuple[int,Any,Any]:
     kwargs.update({
         "stdout"    : PIPE,
         "stderr"    : PIPE,
@@ -880,18 +910,18 @@ def is_systemd_pid1() -> bool:
     if not POSIX:
         return False
     d = load_binary_file("/proc/1/cmdline")
-    return d and d.find(b"/systemd")>=0
+    return d.find(b"/systemd")>=0
 
 
 def get_ssh_port() -> int:
-    #on Linux we can run "ssh -T | grep port"
-    #but this usually requires root permissions to access /etc/ssh/sshd_config
+    # on Linux, we can run "ssh -T | grep port"
+    # but this usually requires root permissions to access /etc/ssh/sshd_config
     if WIN32:
         return 0
     return 22
 
 
-def setuidgid(uid, gid):
+def setuidgid(uid:int, gid:int) -> None:
     if not POSIX:
         return
     log = get_util_logger()
@@ -901,7 +931,7 @@ def setuidgid(uid, gid):
         try:
             username = getpwuid(uid).pw_name
         except KeyError:
-            raise Exception(f"uid {uid} not found") from None
+            raise ValueError(f"uid {uid} not found") from None
         #set the groups:
         if hasattr(os, "initgroups"):   # python >= 2.7
             os.initgroups(username, gid)
@@ -919,7 +949,7 @@ def setuidgid(uid, gid):
             #don't run as root!
             raise
         log.estr(e)
-        log.error(" continuing with gid={os.getgid()}")
+        log.error(f" continuing with gid={os.getgid()}")
     try:
         if os.getuid()!=uid:
             os.setuid(uid)
@@ -932,10 +962,10 @@ def setuidgid(uid, gid):
         log.error(f" continuing with uid={os.getuid()}")
     log(f"new uid={os.getuid()}, gid={os.getgid()}")
 
-def get_peercred(sock):
+def get_peercred(sock) -> Optional[Tuple[int,int,int]]:
+    log = get_util_logger()
     if LINUX:
         SO_PEERCRED = 17
-        log = get_util_logger()
         try:
             creds = sock.getsockopt(socket.SOL_SOCKET, SO_PEERCRED, struct.calcsize(b'3i'))
             pid, uid, gid = struct.unpack(b'3i',creds)
@@ -951,7 +981,7 @@ def get_peercred(sock):
         #then pwd to get the gid?
     return None
 
-def is_socket(sockpath, check_uid=None):
+def is_socket(sockpath:str, check_uid:Optional[int]=None) -> bool:
     try:
         s = os.stat(sockpath)
     except OSError as e:
@@ -960,9 +990,26 @@ def is_socket(sockpath, check_uid=None):
         return False
     if not stat.S_ISSOCK(s.st_mode):
         return False
-    if check_uid is not None:
-        if s.st_uid!=check_uid:
-            #socket uid does not match
-            get_util_logger().debug(f"is_socket({sockpath}) uid {s.st_uid} does not match {check_uid}")
-            return False
+    if check_uid is not None and s.st_uid!=check_uid:
+        #socket uid does not match
+        get_util_logger().debug(f"is_socket({sockpath}) uid {s.st_uid} does not match {check_uid}")
+        return False
     return True
+
+def is_writable(path : str, uid:int=getuid(), gid:int=getgid()) -> bool:
+    if uid==0:
+        return True
+    try:
+        s = os.stat(path)
+    except OSError as e:
+        get_util_logger().debug(f"is_writable({path}) path cannot be accessed: {e}")
+        #socket cannot be accessed
+        return False
+    mode = s.st_mode
+    if s.st_uid==uid and mode & stat.S_IWUSR:
+        #uid has write access
+        return True
+    if s.st_gid==gid and mode & stat.S_IWGRP:
+        #gid has write access:
+        return True
+    return False
