@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -9,19 +9,24 @@
 import os
 import sys
 import types
-import ctypes
-from ctypes import WinDLL, CFUNCTYPE, c_int, POINTER, Structure, byref, sizeof
-from ctypes.wintypes import HWND, DWORD, WPARAM, LPARAM, MSG, POINT, RECT, HGDIOBJ
-from ctypes import CDLL, pythonapi, py_object
+from typing import Callable, List, Tuple, Dict, Any, Type
+from ctypes import (
+    WinDLL, WinError, get_last_error,  # @UnresolvedImport
+    CDLL, pythonapi, py_object,
+    CFUNCTYPE, HRESULT, c_int, c_bool, create_string_buffer, POINTER, Structure, byref, sizeof,  # @UnresolvedImport
+    )
+from ctypes.wintypes import HWND, DWORD, WPARAM, LPARAM, MSG, POINT, RECT, HGDIOBJ, LPCWSTR
 from ctypes.util import find_library
+from gi.repository import GLib
 
-from xpra.client import mixin_features
-from xpra.exit_codes import EXIT_OK
+from xpra.client.gui import mixin_features
+from xpra.exit_codes import ExitCode
+from xpra.common import noop
 from xpra.platform.win32 import constants as win32con, setup_console_event_listener
 from xpra.platform.win32.window_hooks import Win32Hooks
 from xpra.platform.win32.win32_events import KNOWN_EVENTS, POWER_EVENTS
 from xpra.platform.win32.common import (
-    GetSystemMetrics, SetWindowLongW, GetWindowLongW,
+    GetSystemMetrics, SetWindowLongA, GetWindowLongW,
     ClipCursor, GetCursorPos,
     GetDC, ReleaseDC,
     SendMessageA, GetMessageA, TranslateMessage, DispatchMessageA,
@@ -35,9 +40,10 @@ from xpra.platform.win32.common import (
     GetIntSystemParametersInfo,
     GetUserObjectInformationA, OpenInputDesktop, CloseDesktop,
     GetMonitorInfo,
+    GetKeyboardLayoutName,
     )
 from xpra.common import KeyEvent
-from xpra.util import AdHocStruct, csv, envint, envbool, typedict
+from xpra.util import csv, envint, envbool, typedict
 from xpra.os_util import get_util_logger
 from xpra.log import Logger
 
@@ -59,7 +65,11 @@ PyCapsule_GetPointer = pythonapi.PyCapsule_GetPointer
 PyCapsule_GetPointer.restype = HGDIOBJ
 PyCapsule_GetPointer.argtypes = [py_object]
 log("PyCapsute_GetPointer=%s", PyCapsule_GetPointer)
-gdkdll = CDLL(find_library("libgdk-3-0.dll"))
+GDK_DLL_NAME = "libgdk-3-0.dll"
+gdk_dll = find_library(GDK_DLL_NAME)
+if not gdk_dll:
+    raise ImportError(f"ctypes cannot find {GDK_DLL_NAME!r}")
+gdkdll = CDLL(gdk_dll)
 gdk_win32_window_get_handle = gdkdll.gdk_win32_window_get_handle
 gdk_win32_window_get_handle.argtypes = [HGDIOBJ]
 gdk_win32_window_get_handle.restype = HWND
@@ -67,22 +77,24 @@ log("gdkdll=%s", gdkdll)
 
 
 shell32 = WinDLL("shell32", use_last_error=True)
+set_window_group : Callable = noop
 try:
-    from xpra.platform.win32.propsys import set_window_group    #@UnresolvedImport
+    from xpra.platform.win32 import propsys     #@UnresolvedImport
+    set_window_group = propsys.set_window_group
 except ImportError as e:
     log("propsys missing", exc_info=True)
     log.warn("Warning: propsys support missing:")
     log.warn(" %s", e)
     log.warn(" window grouping is not available")
-    set_window_group = None
 
 
+DwmGetWindowAttribute : Callable = noop
 try:
     dwmapi = WinDLL("dwmapi", use_last_error=True)
     DwmGetWindowAttribute = dwmapi.DwmGetWindowAttribute
 except Exception:
     #win XP:
-    DwmGetWindowAttribute = None
+    pass
 
 
 WINDOW_HOOKS = envbool("XPRA_WIN32_WINDOW_HOOKS", True)
@@ -92,6 +104,7 @@ CLIP_CURSOR = WINDOW_HOOKS and envbool("XPRA_WIN32_CLIP_CURSOR", True)
 #GTK3 is fixed, so we don't need this hook:
 MAX_SIZE_HINT = WINDOW_HOOKS and envbool("XPRA_WIN32_MAX_SIZE_HINT", False)
 LANGCHANGE = WINDOW_HOOKS and envbool("XPRA_WIN32_LANGCHANGE", True)
+POLL_LAYOUT = envint("XPRA_WIN32_POLL_LAYOUT", 10)
 
 FORWARD_WINDOWS_KEY = envbool("XPRA_FORWARD_WINDOWS_KEY", True)
 WHEEL = envbool("XPRA_WHEEL", True)
@@ -106,7 +119,7 @@ log("win32 gui settings: FORWARD_WINDOWS_KEY=%s, WHEEL=%s, WHEEL_DELTA=%s",
     FORWARD_WINDOWS_KEY, WHEEL, WHEEL_DELTA)
 
 
-def do_init():
+def do_init() -> None:
     from xpra.platform.win32.dpi import init_dpi
     init_dpi()
     if APP_ID:
@@ -118,9 +131,7 @@ def do_init():
         log.error("Error: failed to setup dwm window color matching", exc_info=True)
 
 
-def init_appid():
-    from ctypes import HRESULT
-    from ctypes.wintypes import LPCWSTR
+def init_appid() -> None:
     SetCurrentProcessExplicitAppUserModelID = shell32.SetCurrentProcessExplicitAppUserModelID
     SetCurrentProcessExplicitAppUserModelID.restype = HRESULT
     SetCurrentProcessExplicitAppUserModelID.argtypes = [LPCWSTR]
@@ -128,7 +139,7 @@ def init_appid():
         log.warn("Warning: failed to set process app ID")
 
 
-def use_stdin():
+def use_stdin() -> bool:
     if os.environ.get("MSYSCON") or os.environ.get("CYGWIN"):
         return False
     stdin = sys.stdin
@@ -146,12 +157,12 @@ def use_stdin():
     return True
 
 
-def get_clipboard_native_class():
+def get_clipboard_native_class() -> str:
     #"xpra.clipboard.translated_clipboard.TranslatedClipboardProtocolHelper"
     return "xpra.platform.win32.clipboard.Win32Clipboard"
 
 
-def get_native_notifier_classes():
+def get_native_notifier_classes() -> List[Type]:
     try:
         from xpra.platform.win32.win32_notifier import Win32_Notifier
         return [Win32_Notifier]
@@ -174,27 +185,27 @@ def get_native_tray_classes():
     return c
 
 def get_native_system_tray_classes(*_args):
-    #Win32Tray cannot set the icon from data
-    #so it cannot be used for application trays
+    # Win32Tray cannot set the icon from data,
+    # so it cannot be used for application trays
     return get_native_tray_classes()
 
-def gl_check():
-    #This is supposed to help py2exe
-    #(must be done after we setup the sys.path in platform.win32.paths):
+def gl_check() -> str:
+    # This is supposed to help `py2exe`
+    # (must be done after we set up the `sys.path` in `platform.win32.paths`):
     try:
         from OpenGL.platform import win32   #@UnresolvedImport @UnusedImport
     except ImportError as e:
         get_util_logger().warn("gl_check()", exc_info=True)
         get_util_logger().warn("Warning: OpenGL bindings are missing")
         get_util_logger().warn(" %s", e)
-        return False
+        return "OpenGL bindings are missing"
     from xpra.platform.win32 import is_wine
     if is_wine():
         return "disabled when running under wine"
-    return None
+    return ""
 
 
-def get_monitor_workarea_for_window(handle):
+def get_monitor_workarea_for_window(handle:int):
     try:
         monitor = MonitorFromWindow(handle, win32con.MONITOR_DEFAULTTONEAREST)
         mi = GetMonitorInfo(monitor)
@@ -216,7 +227,7 @@ def get_monitor_workarea_for_window(handle):
         return None
 
 
-def get_window_handle(window):
+def get_window_handle(window) -> int:
     """ returns the win32 hwnd from a gtk.Window or gdk.Window """
     gdk_window = window
     try:
@@ -230,11 +241,11 @@ def get_window_handle(window):
     #log("get_window_handle(%s) gpointer=%#x, hwnd=%#x", gpointer, hwnd)
     return hwnd
 
-def get_desktop_name():
+def get_desktop_name() -> str:
     try:
         desktop = OpenInputDesktop(0, True, win32con.MAXIMUM_ALLOWED)
         if desktop:
-            buf = ctypes.create_string_buffer(128)
+            buf = create_string_buffer(128)
             r = GetUserObjectInformationA(desktop, win32con.UOI_NAME, buf, len(buf), None)
             if r!=0:
                 desktop_name = buf.value.decode("latin1")
@@ -243,13 +254,13 @@ def get_desktop_name():
     except Exception as e:
         log.warn("Warning: failed to get desktop name")
         log.warn(" %s", e)
-    return None
+    return ""
 
 
-def get_session_type():
+def get_session_type() -> str:
     try:
-        b = ctypes.c_bool()
-        retcode = dwmapi.DwmIsCompositionEnabled(ctypes.byref(b))
+        b = c_bool()
+        retcode = dwmapi.DwmIsCompositionEnabled(byref(b))
         log("get_session_type() DwmIsCompositionEnabled()=%s (retcode=%s)", b.value, retcode)
         if retcode==0 and b.value:
             return "aero"
@@ -296,9 +307,9 @@ def win32_propsys_set_group_leader(self, leader):
     except Exception as e:
         log("set_window_group error", exc_info=True)
         log.error("Error: failed to set group leader")
-        log.error(" %s", e)
+        log.estr(e)
 
-WS_NAMES = {
+WS_NAMES : Dict[int,str] = {
             win32con.WS_BORDER              : "BORDER",
             win32con.WS_CAPTION             : "CAPTION",
             win32con.WS_CHILD               : "CHILD",
@@ -325,10 +336,10 @@ WS_NAMES = {
             win32con.WS_VSCROLL             : "VSCROLL",
             }
 
-def style_str(style):
+def style_str(style) -> str:
     return csv(s for c,s in WS_NAMES.items() if (c & style)==c)
 
-def pointer_grab(window, *args):
+def pointer_grab(window, *args) -> bool:
     hwnd = get_window_handle(window)
     grablog("pointer_grab%s window=%s, hwnd=%s", args, window, hwnd)
     if not hwnd:
@@ -337,7 +348,7 @@ def pointer_grab(window, *args):
     wrect = RECT()
     GetWindowRect(hwnd, byref(wrect))
     grablog("GetWindowRect(%i)=%s", hwnd, wrect)
-    if DwmGetWindowAttribute:
+    if DwmGetWindowAttribute!=noop:
         # Vista & 7 stuff
         rect = RECT()
         DWMWA_EXTENDED_FRAME_BOUNDS = 9
@@ -355,10 +366,10 @@ def pointer_grab(window, *args):
     clip = RECT(*coords)
     r = ClipCursor(clip)
     grablog("ClipCursor%s=%s", coords, r)
-    window._client.pointer_grabbed = window._id
+    window._client.pointer_grabbed = window.wid
     return True
 
-def pointer_ungrab(window, *args):
+def pointer_ungrab(window, *args) -> bool:
     hwnd = get_window_handle(window)
     client = window._client
     grablog("pointer_ungrab%s window=%s, hwnd=%s, pointer_grabbed=%s",
@@ -370,7 +381,7 @@ def pointer_ungrab(window, *args):
     client.pointer_grabbed = None
     return True
 
-def fixup_window_style(self, *_args):
+def fixup_window_style(self, *_args) -> None:
     """ a fixup function we want to call from other places """
     hwnd = get_window_handle(self)
     if not hwnd:
@@ -389,21 +400,15 @@ def fixup_window_style(self, *_args):
             style |= win32con.WS_SYSMENU
             style |= win32con.WS_MAXIMIZEBOX
             style |= win32con.WS_MINIMIZEBOX
-        #we can't enable this because GTK would then get confused
+        #we can't tweak WS_MAXIMIZEBOX and WS_SIZEBOX
+        #to hide the buttons
+        #because GTK would then get confused
         #and paint the window contents at the wrong offset
-        if False:
-            hints = metadata.get("size-constraints")
-            if hints:
-                minw, minh = hints.get("minimum-size", (0, 0))
-                maxw, maxh = hints.get("maximum-size", (0, 0))
-                if minw>0 and minw==maxw and minh>0 and minh==maxh:
-                    #not resizable
-                    style &= ~win32con.WS_MAXIMIZEBOX
-                    style &= ~win32con.WS_SIZEBOX
+        # hints = metadata.get("size-constraints")
         if style!=cur_style:
             log("fixup_window_style() using %s (%#x) instead of %s (%#x) on window %#x with metadata=%s",
                 style_str(style), style, style_str(cur_style), cur_style, hwnd, metadata)
-            SetWindowLongW(hwnd, win32con.GWL_STYLE, style)
+            SetWindowLongA(hwnd, win32con.GWL_STYLE, style)
         else:
             log("fixup_window_style() unchanged style %s (%#x) on window %#x",
                 style_str(style), style, hwnd)
@@ -419,14 +424,14 @@ def fixup_window_style(self, *_args):
             if ws_visible:
                 #with opengl, we need to re-create the window (PITA):
                 if REINIT_VISIBLE_WINDOWS:
-                    client.reinit_window(self._id, self)
+                    client.reinit_window(self.wid, self)
                 self.send_control_refresh(False)
             else:
                 self.send_control_refresh(True)
     except Exception:
         log.warn("failed to fixup window style", exc_info=True)
 
-def set_decorated(self, decorated):
+def set_decorated(self, decorated:bool):
     """ override method which ensures that we call
         fixup_window_style whenever decorations are toggled """
     self.__set_decorated(decorated)         #call the original saved method
@@ -437,7 +442,7 @@ def window_state_updated(window):
     log("window_state_updated(%s)", window)
     fixup_window_style(window)
 
-def apply_maxsize_hints(window, hints):
+def apply_maxsize_hints(window, hints:Dict[str,Any]):
     """ extracts the max-size hints from the hints,
         and passes it to the win32hooks class which can implement it
         (as GTK2 does not honour it properly on win32)
@@ -465,7 +470,7 @@ def apply_maxsize_hints(window, hints):
         else:
             maxw, maxh = workw, workh
     log("apply_maxsize_hints(%s, %s) found min: %sx%s, max: %sx%s", window, hints, minw, minh, maxw, maxh)
-    if (maxw>0 and maxw<32767) or (maxh>0 and maxh<32767):
+    if 0<maxw<32767 or 0<maxh<32767:
         window.win32hooks.max_size = (maxw or 32000), (maxh or 32000)
     elif window.win32hooks.max_size:
         #was set, clear it
@@ -484,17 +489,17 @@ def apply_maxsize_hints(window, hints):
         hints.pop(x, None)
     window_state_updated(window)
 
-def apply_geometry_hints(self, hints):
+def apply_geometry_hints(self, hints:Dict):
     log("apply_geometry_hints(%s)", hints)
     apply_maxsize_hints(self, hints)
     return self.__apply_geometry_hints(hints)   #call the original saved method
 
 def cache_pointer_offset(self, event):
-    #this overrides the window._get_pointer method
-    #so we can cache the GTK position offset for synthetic wheel events
+    # this overrides the `window._get_pointer` method,
+    # so we can cache the GTK position offset for synthetic wheel events
     gtk_x, gtk_y = event.x_root, event.y_root
     pos = POINT()
-    GetCursorPos(ctypes.byref(pos))
+    GetCursorPos(byref(pos))
     x, y = pos.x, pos.y
     self.win32_pointer_offset = gtk_x-x, gtk_y-y
     return gtk_x, gtk_y
@@ -503,7 +508,7 @@ def cache_pointer_offset(self, event):
 def no_set_group(*_args):
     """ provide a dummy implementation """
 
-def add_window_hooks(window):
+def add_window_hooks(window) -> None:
     log("add_window_hooks(%s) WINDOW_HOOKS=%s, GROUP_LEADER=%s, UNDECORATED_STYLE=%s",
             window, WINDOW_HOOKS, GROUP_LEADER, UNDECORATED_STYLE)
     log(" MAX_SIZE_HINT=%s, MAX_SIZE_HINT=%s", MAX_SIZE_HINT, MAX_SIZE_HINT)
@@ -526,24 +531,23 @@ def add_window_hooks(window):
     log("add_window_hooks(%s) gdk window=%s, hwnd=%#x", window, gdk_window, handle)
 
     if GROUP_LEADER:
-        #windows 7 onwards can use AppUserModel to emulate the group leader stuff:
+        # MSWindows 7 onwards can use AppUserModel to emulate the group leader stuff:
         log("win32 hooks: set_window_group=%s", set_window_group)
-        if set_window_group:
-            gdk_window.set_group = types.MethodType(win32_propsys_set_group_leader, gdk_window)
-            log("hooked group leader override using %s", set_window_group)
+        gdk_window.set_group = types.MethodType(win32_propsys_set_group_leader, gdk_window)
+        log("hooked group leader override using %s", set_window_group)
 
     if UNDECORATED_STYLE:
         #OR windows never have any decorations or taskbar menu
         if not window._override_redirect:
-            #the method to call to fix things up:
+            # the method to call to fix things up:
             window.fixup_window_style = types.MethodType(fixup_window_style, window)
-            #override set_decorated so we can preserve the taskbar menu for undecorated windows
+            # override `set_decorated` so we can preserve the taskbar menu for undecorated windows
             window.__set_decorated = window.set_decorated
             window.set_decorated = types.MethodType(set_decorated, window)
-            #override after_window_state_updated so we can re-add the missing style options
-            #(somehow doing it from on_realize which calls add_window_hooks is not enough)
+            # override `after_window_state_updated` so we can re-add the missing style options
+            # (somehow doing it from on_realize which calls add_window_hooks is not enough)
             window.connect("state-updated", window_state_updated)
-            #call it at least once:
+            # call it at least once:
             window.fixup_window_style()
 
     if CLIP_CURSOR:
@@ -569,7 +573,7 @@ def add_window_hooks(window):
 
         if LANGCHANGE:
             def inputlangchange(_hwnd, _event, wParam, lParam):
-                log("WM_INPUTLANGCHANGE: character set: %i, input locale identifier: %i", wParam, lParam)
+                keylog("WM_INPUTLANGCHANGE: character set: %i, input locale identifier: %i", wParam, lParam)
                 window.keyboard_layout_changed("WM_INPUTLANGCHANGE", wParam, lParam)
             win32hooks.add_window_event_handler(win32con.WM_INPUTLANGCHANGE, inputlangchange)
 
@@ -597,7 +601,8 @@ def add_window_hooks(window):
                         deltax = units
                         deltay = 0
                     pointer = window.get_mouse_position()
-                    client.wheel_event(wid, deltax, deltay, pointer)
+                    device_id = -1
+                    client.wheel_event(device_id, wid, deltax, deltay, pointer)
             def mousewheel(_hwnd, _event, wParam, lParam):
                 handle_wheel(VERTICAL, wParam, lParam)
                 return 0
@@ -605,9 +610,10 @@ def add_window_hooks(window):
                 handle_wheel(HORIZONTAL, wParam, lParam)
                 return 0
             win32hooks.add_window_event_handler(win32con.WM_MOUSEWHEEL, mousewheel)
+            win32hooks.add_window_event_handler(win32con.WM_MOUSEHWHEEL, mousehwheel)
 
 
-def remove_window_hooks(window):
+def remove_window_hooks(window) -> None:
     try:
         win32hooks = getattr(window, "win32hooks", None)
         if win32hooks:
@@ -618,14 +624,14 @@ def remove_window_hooks(window):
         log.error("remove_window_hooks(%s)", exc_info=True)
 
 
-def get_xdpi():
+def get_xdpi() -> int:
     try:
         return _get_device_caps(win32con.LOGPIXELSX)
     except Exception as e:
         log.warn("failed to get xdpi: %s", e)
     return -1
 
-def get_ydpi():
+def get_ydpi() -> int:
     try:
         return _get_device_caps(win32con.LOGPIXELSY)
     except Exception as e:
@@ -643,32 +649,32 @@ FE_FONTSMOOTHINGORIENTATIONVRGB = 0x0003
 SPI_GETFONTSMOOTHINGTYPE        = 0x200A
 FE_FONTSMOOTHINGCLEARTYPE       = 0x0002
 FE_FONTSMOOTHINGDOCKING         = 0x8000
-FE_ORIENTATION_STR = {
+FE_ORIENTATION_STR : Dict[int,str] = {
                       FE_FONTSMOOTHINGORIENTATIONBGR    : "BGR",
                       FE_FONTSMOOTHINGORIENTATIONRGB    : "RGB",
                       FE_FONTSMOOTHINGORIENTATIONVBGR   : "VBGR",
                       FE_FONTSMOOTHINGORIENTATIONVRGB   : "VRGB",
                       }
-FE_FONTSMOOTHING_STR = {
+FE_FONTSMOOTHING_STR : Dict[int,str] = {
     0                           : "Normal",
     FE_FONTSMOOTHINGCLEARTYPE   : "ClearType",
     }
 
 
-def _add_SPI(info, constant, name, convert, default=None):
+def _add_SPI(info:Dict[str,Any], constant:int, name:str, convert:Callable, default:Any=None) -> None:
     v = GetIntSystemParametersInfo(constant)
     if v is not None:
         info[name] = convert(v)
     elif default is not None:
         info[name] = default
 
-def get_antialias_info():
-    info = {}
+def get_antialias_info() -> Dict[str,Any]:
+    info : Dict[str,Any] = {}
     try:
         _add_SPI(info, SPI_GETFONTSMOOTHING, "enabled", bool)
         #"Valid contrast values are from 1000 to 2200. The default value is 1400."
         _add_SPI(info, SPI_GETFONTSMOOTHINGCONTRAST, "contrast", int)
-        def orientation(v):
+        def orientation(v) -> str:
             return FE_ORIENTATION_STR.get(v, "unknown")
         _add_SPI(info, SPI_GETFONTSMOOTHINGORIENTATION, "orientation", orientation)
         def smoothing_type(v):
@@ -679,7 +685,7 @@ def get_antialias_info():
         log.warn("failed to query antialias info: %s", e)
     return info
 
-def get_mouse_config():
+def get_mouse_config() -> Dict[str,Any]:
     #not all are present in win32con?
     SM_CMOUSEBUTTONS = 43
     SM_CXDRAG = 68
@@ -698,7 +704,7 @@ def get_mouse_config():
     #rate for each direction:
     _add_SPI(wheel_info, SPI_GETWHEELSCROLLLINES, "lines", int, 3)
     _add_SPI(wheel_info, SPI_GETWHEELSCROLLCHARS, "chars", int, 3)
-    info = {
+    info : Dict[str,Any] = {
             "present"       : bool(GetSystemMetrics(SM_MOUSEPRESENT)),
             "wheel"         : wheel_info,
             "buttons"       : GetSystemMetrics(SM_CMOUSEBUTTONS),
@@ -756,9 +762,9 @@ def get_workarea():
 #ie: for a 60 pixel bottom bar on the second monitor at 1280x800:
 # [(0,0,1920,1080), (0,0,1280,740)]
 MONITORINFOF_PRIMARY = 1
-def get_workareas():
+def get_workareas() -> List[Tuple[int,int,int,int]]:
     try:
-        workareas = []
+        workareas : List[Tuple[int,int,int,int]] = []
         for m in EnumDisplayMonitors():
             mi = GetMonitorInfo(m)
             screenlog("get_workareas() GetMonitorInfo(%s)=%s", m, mi)
@@ -803,6 +809,10 @@ def get_vrefresh():
         log.warn("Warning: failed to query the display vertical refresh rate:")
         log.warn(" %s", e)
         v = -1
+    if v in (0, 1):
+        # as per the docs:
+        # "A vertical refresh rate value of 0 or 1 represents the display hardware's default refresh rate"
+        return -1
     screenlog("get_vrefresh()=%s", v)
     return v
 
@@ -952,6 +962,7 @@ def getTaskbar():
     # pylint: disable=import-outside-toplevel
     global TaskbarLib
     if TaskbarLib is None:
+        taskbar_tlb = ""
         try:
             from xpra.platform.win32.comtypes_util import QuietenLogging, find_tlb_file, comtypes_init
             taskbar_tlb = find_tlb_file("TaskbarLib.tlb")
@@ -979,7 +990,7 @@ def getTaskbar():
     return taskbar
 
 
-def set_window_progress(window, pct):
+def set_window_progress(window, pct:int):
     taskbar = getattr(window, "taskbar", None)
     if not taskbar:
         taskbar = getTaskbar()
@@ -999,7 +1010,7 @@ WTS_SESSION_LOGOFF          = 0x6
 WTS_SESSION_LOCK            = 0x7
 WTS_SESSION_UNLOCK          = 0x8
 WTS_SESSION_REMOTE_CONTROL  = 0x9
-WTS_SESSION_EVENTS = {
+WTS_SESSION_EVENTS : Dict[int, str] = {
                       WTS_CONSOLE_CONNECT       : "CONSOLE CONNECT",
                       WTS_CONSOLE_DISCONNECT    : "CONSOLE_DISCONNECT",
                       WTS_REMOTE_CONNECT        : "REMOTE_CONNECT",
@@ -1047,16 +1058,18 @@ class ClientExtras:
                 el.add_event_callback(win32con.WM_ENDSESSION,       self.end_session)
         except Exception as e:
             log.error("Error: cannot register focus and power callbacks:")
-            log.error(" %s", e)
+            log.estr(e)
         self.keyboard_hook_id = None
+        self.keyboard_poll_timer : int = 0
+        self.keyboard_id : int = self.get_keyboard_layout_id()
         if FORWARD_WINDOWS_KEY and mixin_features.windows:
             from xpra.make_thread import start_thread
             start_thread(self.init_keyboard_listener, "keyboard-listener", daemon=True)
 
-    def ready(self):
+    def ready(self) -> None:
         """ nothing specific to do here """
 
-    def cleanup(self):
+    def cleanup(self) -> None:
         log("ClientExtras.cleanup()")
         self._exit = True
         cha = self._console_handler_added
@@ -1072,6 +1085,10 @@ class ClientExtras:
         if khid:
             self.keyboard_hook_id = None
             UnhookWindowsHookEx(khid)
+        kpt = self.keyboard_poll_timer
+        if kpt:
+            self.keyboard_poll_timer = 0
+            GLib.source_remove(kpt)
         sst = self._screensaver_timer
         if sst:
             self._screensaver_timer = 0
@@ -1079,12 +1096,33 @@ class ClientExtras:
         log("ClientExtras.cleanup() ended")
         #self.client = None
 
-    def init_keyboard_listener(self):
+    def get_keyboard_layout_id(self) -> int:
+        name_buf = create_string_buffer(win32con.KL_NAMELENGTH)
+        if not GetKeyboardLayoutName(name_buf):
+            return 0
+        log(f"layout-name={name_buf.value}")
+        try:
+            #win32 API returns a hex string
+            return int(name_buf.value, 16)
+        except ValueError:
+            log.warn("Warning: failed to parse keyboard layout code '%s'", name_buf.value)
+        return 0
+
+    def poll_layout(self) -> None:
+        self.keyboard_poll_timer = 0
+        klid = self.get_keyboard_layout_id()
+        if klid and klid!=self.keyboard_id:
+            self.keyboard_id = klid
+            self.client.window_keyboard_layout_changed()
+
+    def init_keyboard_listener(self) -> None:
         class KBDLLHOOKSTRUCT(Structure):
             _fields_ = [("vk_code", DWORD),
                         ("scan_code", DWORD),
                         ("flags", DWORD),
                         ("time", c_int),]
+            def toString(self):
+                return f"KBDLLHOOKSTRUCT({self.vk_code}, {self.scan_code}, {self.flags:x}, {self.time})"
         DOWN = [win32con.WM_KEYDOWN, win32con.WM_SYSKEYDOWN]
         #UP = [win32con.WM_KEYUP, win32con.WM_SYSKEYUP]
         ALL_KEY_EVENTS = {win32con.WM_KEYDOWN       : "KEYDOWN",
@@ -1092,8 +1130,15 @@ class ClientExtras:
                           win32con.WM_KEYUP         : "KEYUP",
                           win32con.WM_SYSKEYUP      : "SYSKEYUP",
                           }
-        def low_level_keyboard_handler(nCode, wParam, lParam):
+        def low_level_keyboard_handler(nCode:int, wParam:int, lParam:int):
             log("WH_KEYBOARD_LL: %s", (nCode, wParam, lParam))
+            kh = getattr(self.client, "keyboard_helper", None)
+            locked = getattr(kh, "locked", False)
+            if POLL_LAYOUT and self.keyboard_poll_timer==0 and not locked:
+                self.keyboard_poll_timer = GLib.timeout_add(POLL_LAYOUT, self.poll_layout)
+            if nCode<0:
+                #docs say we should not process this event:
+                return CallNextHookEx(0, nCode, wParam, lParam)
             try:
                 scan_code = lParam.contents.scan_code
                 vk_code = lParam.contents.vk_code
@@ -1104,7 +1149,7 @@ class ClientExtras:
                            win32con.VK_RWIN   : "Super_R",
                            win32con.VK_TAB    : "Tab",
                            }.get(vk_code)
-                modifiers = []
+                modifiers : List[str] = []
                 kh = self.client.keyboard_helper
                 key_event_type = ALL_KEY_EVENTS.get(wParam)
                 #log("low_level_keyboard_handler(%s, %s, %s) vk_code=%i, scan_code=%i, keyname=%s, key_event_type=%s, focused=%s, keyboard_grabbed=%s", nCode, wParam, lParam, vk_code, scan_code, keyname, key_event_type, focused, self.client.keyboard_grabbed)
@@ -1156,7 +1201,7 @@ class ClientExtras:
                         return 1
             except Exception as e:
                 keylog.error("Error: low level keyboard hook failed")
-                keylog.error(" %s", e)
+                keylog.estr(e)
             return CallNextHookEx(0, nCode, wParam, lParam)
         # Our low level handler signature.
         CMPFUNC = CFUNCTYPE(c_int, WPARAM, LPARAM, POINTER(KBDLLHOOKSTRUCT))
@@ -1172,7 +1217,7 @@ class ClientExtras:
             ret = GetMessageA(lpmsg, 0, 0, 0)
             keylog("keyboard listener: GetMessage()=%s", ret)
             if ret==-1:
-                raise ctypes.WinError(ctypes.get_last_error())
+                raise WinError(get_last_error())
             if ret==0:
                 keylog("GetMessage()=0, exiting loop")
                 return
@@ -1182,7 +1227,7 @@ class ClientExtras:
             keylog("DispatchMessageA(%#x)=%s", lpmsg, r)
 
 
-    def wm_move(self, wParam, lParam):
+    def wm_move(self, wParam:int, lParam:int) -> None:
         c = self.client
         log("WM_MOVE: %s/%s client=%s", wParam, lParam, c)
         if c:
@@ -1190,7 +1235,7 @@ class ClientExtras:
             #but we do want to process it as such (see window reinit code)
             c.screen_size_changed()
 
-    def end_session(self, wParam, lParam):
+    def end_session(self, wParam:int, lParam:int):
         log("WM_ENDSESSION")
         c = self.client
         if not c:
@@ -1198,7 +1243,6 @@ class ClientExtras:
         ENDSESSION_CLOSEAPP = 0x1
         ENDSESSION_CRITICAL = 0x40000000
         ENDSESSION_LOGOFF   = 0x80000000
-        reason = None
         if (wParam & ENDSESSION_CLOSEAPP) and wParam:
             reason = "restart manager request"
         elif wParam & ENDSESSION_CRITICAL:
@@ -1207,10 +1251,10 @@ class ClientExtras:
             reason = "logoff"
         else:
             return
-        c.disconnect_and_quit(EXIT_OK, reason)
+        c.disconnect_and_quit(ExitCode.OK, reason)
 
-    def session_change_event(self, event, session):
-        event_name = WTS_SESSION_EVENTS.get(event, event)
+    def session_change_event(self, event:int, session:int):
+        event_name = WTS_SESSION_EVENTS.get(event) or str(event)
         log("WM_WTSSESSION_CHANGE: %s on session %#x", event_name, session)
         c = self.client
         if not c:
@@ -1222,14 +1266,14 @@ class ClientExtras:
             log("will unfreeze all the windows")
             #don't unfreeze directly from here,
             #as the system may not be fully usable yet (see #997)
-            from gi.repository import GLib
             GLib.idle_add(c.unfreeze)
 
 
-    def inputlangchange(self, wParam, lParam):
-        log("WM_INPUTLANGCHANGE: %i, %i", wParam, lParam)
+    def inputlangchange(self, wParam:int, lParam:int):
+        keylog("WM_INPUTLANGCHANGE: %i, %i", wParam, lParam)
+        self.poll_layout()
 
-    def inichange(self, wParam, lParam):
+    def inichange(self, wParam:int, lParam:int):
         if lParam:
             from ctypes import c_char_p
             log("WM_WININICHANGE: %#x=%s", lParam, c_char_p(lParam).value)
@@ -1237,7 +1281,7 @@ class ClientExtras:
             log("WM_WININICHANGE: %i, %i", wParam, lParam)
 
 
-    def activateapp(self, wParam, lParam):
+    def activateapp(self, wParam:int, lParam:int) -> None:
         c = self.client
         log("WM_ACTIVATEAPP: %s/%s client=%s", wParam, lParam, c)
         if not c:
@@ -1254,7 +1298,7 @@ class ClientExtras:
                 fixup_window_style()
 
 
-    def power_broadcast_event(self, wParam, lParam):
+    def power_broadcast_event(self, wParam:int, lParam:int):
         c = self.client
         log("WM_POWERBROADCAST: %s/%s client=%s", POWER_EVENTS.get(wParam, wParam), lParam, c)
         #maybe also "PBT_APMQUERYSUSPEND" and "PBT_APMQUERYSTANDBY"?
@@ -1266,7 +1310,7 @@ class ClientExtras:
             c.resume()
 
 
-    def handle_console_event(self, event):
+    def handle_console_event(self, event:int) -> int:
         c = self.client
         event_name = KNOWN_EVENTS.get(event, event)
         log("handle_console_event(%s) client=%s, event_name=%s", event, c, event_name)
@@ -1279,15 +1323,13 @@ class ClientExtras:
             log.info("received console event %s", str(event_name).replace("_EVENT", ""))
         else:
             log.warn("unknown console event: %s", event_name)
-        if event==win32con.CTRL_C_EVENT:
-            if c:
-                log("calling=%s", c.signal_disconnect_and_quit)
-                c.signal_disconnect_and_quit(0, "CTRL_C")
-                return 1
-        if event==win32con.CTRL_CLOSE_EVENT:
-            if c:
-                c.signal_disconnect_and_quit(0, "CTRL_CLOSE")
-                return 1
+        if c and event==win32con.CTRL_C_EVENT:
+            log("calling=%s", c.signal_disconnect_and_quit)
+            c.signal_disconnect_and_quit(0, "CTRL_C")
+            return 1
+        if c and event==win32con.CTRL_CLOSE_EVENT:
+            c.signal_disconnect_and_quit(0, "CTRL_CLOSE")
+            return 1
         return 0
 
 
@@ -1299,25 +1341,11 @@ def main():
             log.enable_debug()
             win32_event_logger.enable_debug()
 
-        from gi.repository import GLib
-
         log.info("Event loop is running")
         loop = GLib.MainLoop()
 
-        def suspend():
-            log.info("suspend event")
-        def resume():
-            log.info("resume event")
-        fake_client = AdHocStruct()
-        fake_client._focused = False
-        fake_client.keyboard_grabbed = False
-        fake_client.window_with_grab = None
-        fake_client.suspend = suspend
-        fake_client.resume = resume
-        fake_client.keyboard_helper = None
-        def noop(*args):
-            pass
-        fake_client.timeout_add = noop
+        from xpra.client.gui.fake_client import FakeClient
+        fake_client = FakeClient()
         def signal_quit(*_args):
             loop.quit()
         fake_client.signal_disconnect_and_quit = signal_quit
