@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (c) 2017-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (c) 2017-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -9,7 +9,9 @@
 from ctypes.wintypes import HANDLE, DWORD
 from ctypes import byref, sizeof, create_string_buffer, cast, c_void_p, c_long, pointer, POINTER
 from threading import Thread
+from typing import Callable, Optional
 
+from xpra.common import  noop
 from xpra.log import Logger
 from xpra.util import envbool
 from xpra.os_util import strtobytes
@@ -77,17 +79,17 @@ GENERIC_ALL = 0x10000000
 
 
 class NamedPipeListener(Thread):
-    def __init__(self, pipe_name, new_connection_cb=None):
+    def __init__(self, pipe_name:str, new_connection_cb:Callable=noop):
         log("NamedPipeListener(%s, %s)", pipe_name, new_connection_cb)
         self.pipe_name = pipe_name
-        self.new_connection_cb = new_connection_cb or self.new_connection
+        if new_connection_cb!=noop:
+            self.new_connection_cb = new_connection_cb
         self.exit_loop = False
         super().__init__(name="NamedPipeListener-%s" % pipe_name)
         self.daemon = True
-        self.security_attributes = None
-        self.security_descriptor = None
+        self.security_attributes : Optional[SECURITY_ATTRIBUTES] = None
+        self.security_descriptor : Optional[SECURITY_DESCRIPTOR] = None
         self.token_process = HANDLE()
-        self.tokens = []
         cur_proc = GetCurrentProcess()
         log("GetCurrentProcess()=%#x", cur_proc)
         TOKEN_QUERY = 0x8
@@ -98,11 +100,11 @@ class NamedPipeListener(Thread):
     def __repr__(self):
         return "NamedPipeListener(%s)" % self.pipe_name
 
-    def stop(self):
+    def stop(self) -> None:
         log("%s.stop()", self)
         self.exit_loop = True
 
-    def run(self):
+    def run(self) -> None:
         log("%s.run()", self)
         try:
             self.do_run()
@@ -110,13 +112,12 @@ class NamedPipeListener(Thread):
             log.error("Error: named pipe '%s'", self.pipe_name, exc_info=True)
         tp = self.token_process
         if tp:
-            self.token_process = None
+            self.token_process = HANDLE(0)
             CloseHandle(tp)
-        self.tokens = []
         self.security_attributes = None
         self.security_descriptor = None
 
-    def do_run(self):
+    def do_run(self) -> None:
         pipe_handle = None
         while not self.exit_loop:
             if not pipe_handle:
@@ -126,13 +127,13 @@ class NamedPipeListener(Thread):
                     log("CreatePipeHandle()", exc_info=True)
                     log.error("Error: failed to create named pipe")
                     log.error(" at path '%s'", self.pipe_name)
-                    log.error(" %s", e)
+                    log.estr(e)
                     return
                 log("CreatePipeHandle()=%#x", pipe_handle)
                 if c_long(pipe_handle).value==INVALID_HANDLE_VALUE:
                     log.error("Error: invalid handle for named pipe '%s'", self.pipe_name)
-                    e = GetLastError()
-                    log.error(" '%s' (%i)", FormatMessageSystem(e).rstrip("\n\r."), e)
+                    err : int = GetLastError()
+                    log.error(" '%s' (%i)", FormatMessageSystem(err).rstrip("\n\r."), err)
                     return
             event = CreateEventA(None, True, False, None)
             overlapped = OVERLAPPED()
@@ -145,11 +146,11 @@ class NamedPipeListener(Thread):
             if self.exit_loop:
                 break
             if r==0:
-                e = GetLastError()
-                log("GetLastError()=%s (%i)", FormatMessageSystem(e).rstrip("\n\r"), e)
-                if e==ERROR_PIPE_CONNECTED:
-                    pass
-                elif e==ERROR_IO_PENDING:
+                err = GetLastError()
+                log("GetLastError()=%s (%i)", FormatMessageSystem(err).rstrip("\n\r"), err)
+                if err==ERROR_PIPE_CONNECTED:
+                    "non-zero, but OK!"
+                elif err==ERROR_IO_PENDING:
                     while not self.exit_loop:
                         r = WaitForSingleObject(event, INFINITE)
                         log("WaitForSingleObject(..)=%s", WAIT_STR.get(r, r))
@@ -164,7 +165,7 @@ class NamedPipeListener(Thread):
                         break
                 else:
                     log.error("Error: cannot connect to named pipe '%s'", self.pipe_name)
-                    log.error(" error %s", e)
+                    log.error(" error %s", err)
                     CloseHandle(pipe_handle)
                     pipe_handle = None
                 if self.exit_loop:
@@ -176,18 +177,18 @@ class NamedPipeListener(Thread):
         if pipe_handle:
             self.close_handle(pipe_handle)
 
-    def close_handle(self, pipe_handle):
+    def close_handle(self, pipe_handle:HANDLE) -> None:
         try:
             log("CloseHandle(%#x)", pipe_handle)
             CloseHandle(pipe_handle)
         except Exception:
             log("CloseHandle(%#x)", pipe_handle, exc_info=True)
 
-    def new_connection(self, listener, pipe_handle):
-        log.info("new_connection(%s, %#x)", listener, pipe_handle)
+    def new_connection(self, socktype, listener, pipe_handle:HANDLE) -> None:
+        log.info("new_connection(%s, %s, %#x)", socktype, listener, pipe_handle)
         self.close_handle(pipe_handle)
 
-    def CreatePipeHandle(self):
+    def CreatePipeHandle(self) -> HANDLE:
         sa = self.CreatePipeSecurityAttributes()
         log("CreateNamedPipeA using %s (UNRESTRICTED=%s)", sa, UNRESTRICTED)
         return CreateNamedPipeA(strtobytes(self.pipe_name), PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
@@ -203,13 +204,12 @@ class NamedPipeListener(Thread):
                 raise WindowsError()  # @UndefinedVariable
         log("GetTokenInformation data size %#x", data_size.value)
         token_data = create_string_buffer(data_size.value)
-        self.tokens.append(token_data)
         if not GetTokenInformation(self.token_process, token_type, byref(token_data), data_size.value, byref(data_size)):
             raise WindowsError()    #@UndefinedVariable
         token = cast(token_data, POINTER(token_struct)).contents
         return token
 
-    def CreatePipeSecurityAttributes(self):
+    def CreatePipeSecurityAttributes(self) -> SECURITY_ATTRIBUTES:
         user = self.GetToken(TokenUser, TOKEN_USER)
         user_SID = user.SID.contents
         log("user SID=%s, attributes=%#x", user_SID, user.ATTRIBUTES)
@@ -303,7 +303,6 @@ def main():
     import sys
     for verbose in ("-v", "--verbose"):
         if verbose in sys.argv:
-            sys.argv.pop(verbose)
             log.enable_debug()
     pipe_name = "Xpra\\Test"
     if len(sys.argv)>1:
