@@ -1,13 +1,14 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2023 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 from math import sqrt
 from time import monotonic
+from typing import Dict, Any, Tuple, List
 from collections import deque
 
 from xpra.server.cystats import (                                           #@UnresolvedImport
@@ -22,7 +23,7 @@ log = Logger("network", "stats")
 NRECS = 500
 
 
-def safeint(v, default=0):
+def safeint(v, default : int=0) -> int:
     try:
         return int(v)
     except ValueError:
@@ -39,12 +40,12 @@ class GlobalPerformanceStatistics:
     DEFAULT_LATENCY = 0.1
 
     def reset(self, maxlen=NRECS):
-        def d(maxlen=maxlen):
+        def d(maxlen=maxlen) -> deque:
             return deque(maxlen=maxlen)
         # mmap state:
         self.mmap_size = 0
         self.mmap_bytes_sent = 0
-        self.mmap_free_size = 0                             #how much of the mmap space is left (may be negative if we failed to write the last chunk)
+        self.mmap_free_size = 0                             #how much of the mmap space is left (it may be negative if we failed to write the last chunk)
         # queue statistics:
         self.compression_work_qsizes = d()                  #size of the compression_work_queue before we add a new record to it
                                                             #(event_time, size)
@@ -95,7 +96,7 @@ class GlobalPerformanceStatistics:
 
     def record_latency(self, wid : int, damage_packet_sequence : int, decode_time : int,
                        start_send_at, end_send_at,
-                       pixels : int, bytecount : int, latency : int):
+                       pixels : int, bytecount : int, latency : int) -> None:
         now = monotonic()
         send_time = end_send_at-start_send_at   #how long we spend calling socket.send
         total = now-start_send_at               #from the moment we start sending until we process the ack here
@@ -113,11 +114,11 @@ class GlobalPerformanceStatistics:
         self.client_latency.append((wid, now, pixels, net_total_latency))
         self.frame_total_latency.append((wid, now, pixels, latency))
 
-    def get_damage_pixels(self, wid):
-        """ returns the list of (event_time, pixelcount) for the given window id """
-        return [(event_time, value) for event_time, dwid, value in tuple(self.damage_packet_qpixels) if dwid==wid]
+    def get_damage_pixels(self, wid:int) -> Tuple[Tuple[float,int],...]:
+        """ returns the tuple of (event_time, pixelcount) for the given window id """
+        return tuple((event_time, value) for event_time, dwid, value in tuple(self.damage_packet_qpixels) if dwid==wid)
 
-    def update_averages(self):
+    def update_averages(self) -> None:
         def latency_averages(values):
             avg, recent = calculate_time_weighted_average(values)
             return max(0.001, avg), max(0.001, recent)
@@ -153,12 +154,11 @@ class GlobalPerformanceStatistics:
         self.avg_congestion_send_speed = int(acss)
         #how often we get congestion events:
         #first chunk it into second intervals
-        min_time = now-10
         cst = tuple(x[0] for x in css)
         cps = []
         for t in range(10):
             etime = now-t
-            matches = tuple(1 for x in cst if x>etime-1 and x<=etime) or (0,)
+            matches = tuple(1 for x in cst if etime-1<x<=etime) or (0,)
             cps.append((etime, sum(matches)))
         #log("cps(%s)=%s (now=%s)", cst, cps, now)
         self.congestion_value = time_weighted_average(cps)
@@ -168,9 +168,9 @@ class GlobalPerformanceStatistics:
             #(wid, event_time, no_of_pixels, latency)
             self.avg_frame_total_latency = safeint(calculate_size_weighted_average(edata)[1])
 
-    def get_factors(self, pixel_count):
+    def get_factors(self, pixel_count:int) -> List[Tuple[str,Dict,float,float]]:
         factors = []
-        def mayaddfac(metric, info, factor, weight):
+        def mayaddfac(metric:str, info:Dict, factor:float, weight:float):
             if weight>0.01:
                 factors.append((metric, info, factor, weight))
         if self.client_latency:
@@ -205,12 +205,12 @@ class GlobalPerformanceStatistics:
             #full: effective range is 0.0 to ~1.2
             full = 1.0-self.mmap_free_size/self.mmap_size
             #aim for ~33%
-            mayaddfac("mmap-area", "%s%% full" % int(100*full), logp(3*full), (3*full)**2)
+            mayaddfac("mmap-area", {"full-pct" : int(100*full)}, logp(3*full), (3*full)**2)
         if self.congestion_value>0:
             mayaddfac("congestion", {}, 1+self.congestion_value, self.congestion_value*10)
         return factors
 
-    def get_connection_info(self) -> dict:
+    def get_connection_info(self) -> Dict[str,Any]:
         latencies = tuple(int(x*1000) for (_, _, _, x) in tuple(self.client_latency))
         now = monotonic()
         info = {
@@ -232,14 +232,14 @@ class GlobalPerformanceStatistics:
         return info
 
 
-    def get_info(self) -> dict:
+    def get_info(self) -> Dict[str,Any]:
         cwqsizes = tuple(x[1] for x in tuple(self.compression_work_qsizes))
         pqsizes = tuple(x[1] for x in tuple(self.packet_qsizes))
         now = monotonic()
         time_limit = now-60             #ignore old records (60s)
         client_latency = max(0, self.avg_frame_total_latency-
                              int((self.avg_client_ping_latency+self.avg_server_ping_latency)//2))
-        info = {
+        info : Dict[str,Any] = {
             "damage" : {
                 "events"        : self.damage_events_count,
                 "packets_sent"  : self.packet_count,
@@ -252,15 +252,18 @@ class GlobalPerformanceStatistics:
                 "frame-total-latency" : self.avg_frame_total_latency,
                 "client-latency"    : client_latency,
                 },
-            "encoding" : {"decode_errors"   : self.decode_errors},
             "connection" : self.get_connection_info(),
             }
+        einfo : Dict[str,Any] = {
+            "decode_errors" : self.decode_errors,
+        }
+        info["encoding"] = einfo
         if self.quality:
             ql = tuple(quality for _,_,quality in tuple(self.quality))
-            info["encoding"]["quality"] = get_list_stats(ql)
+            einfo["quality"] = get_list_stats(ql)
         if self.speed:
             sl = tuple(speed for _,_,speed in tuple(self.speed))
-            info["encoding"]["speed"] = get_list_stats(sl)
+            einfo["speed"] = get_list_stats(sl)
         #client pixels per second:
         #pixels per second: decode time and overall
         total_pixels = 0                #total number of pixels processed
@@ -279,13 +282,15 @@ class GlobalPerformanceStatistics:
         log("total_time=%s, total_pixels=%s", total_time, total_pixels)
         if total_time>0:
             pixels_decoded_per_second = int(total_pixels *1000*1000 / total_time)
-            info["encoding"]["pixels_decoded_per_second"] = pixels_decoded_per_second
+            einfo["pixels_decoded_per_second"] = pixels_decoded_per_second
         if start_time:
             elapsed = now-start_time
-            pixels_per_second = int(total_pixels/elapsed)
-            info.setdefault("encoding", {}).update({
-                                                    "pixels_per_second"     : pixels_per_second,
-                                                    "regions_per_second"    : int(len(region_sizes)/elapsed),
-                                                    "average_region_size"   : int(total_pixels/len(region_sizes)),
-                                                    })
+            if elapsed:
+                pixels_per_second = int(total_pixels/elapsed)
+                einfo.update({
+                    "pixels_per_second"     : pixels_per_second,
+                    "regions_per_second"    : int(len(region_sizes)/elapsed),
+                })
+            if region_sizes:
+                einfo["average_region_size"] = int(total_pixels/len(region_sizes))
         return info
