@@ -1,14 +1,15 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2017-2022 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2017-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import re
 import os.path
+from typing import Optional, Dict, Callable, Any, Tuple
 
 from xpra.util import ellipsizer, envbool
-from xpra.os_util import getuid, OSX, POSIX, bytestostr, get_proc_cmdline
+from xpra.os_util import getuid, OSX, POSIX, get_proc_cmdline
 from xpra.platform.paths import get_user_conf_dirs, get_system_conf_dirs
 from xpra.log import Logger
 
@@ -27,12 +28,13 @@ def getprop(window, prop):
         return window.get_property(prop)
     except TypeError:
         log.error("Error querying %s on %s", prop, window, exc_info=True)
+        return None
 
 ################################################################
 # generic file parsing functions
 ################################################################
 
-def _load_dict_file(filename, parser) -> dict:
+def _load_dict_file(filename:str, parser:Callable) -> Dict:
     #filter out comments and remove line endings
     lines = []
     with open(filename, "r", encoding="utf8") as f:
@@ -44,7 +46,7 @@ def _load_dict_file(filename, parser) -> dict:
     log("_load_dict_file(%s)=%s", filename, ellipsizer(lines))
     return parser(lines)
 
-def _load_dict_dir(d, parser) -> dict:
+def _load_dict_dir(d:str, parser:Callable) -> Dict:
     #load all the .conf files from the directory
     if not os.path.exists(d) or not os.path.isdir(d):
         log("load_content_categories_dir(%s) directory not found", d)
@@ -59,11 +61,11 @@ def _load_dict_dir(d, parser) -> dict:
                 except Exception as e:
                     log("_load_dict_dir(%s)", cc_file, exc_info=True)
                     log.error("Error loading file data from '%s'", cc_file)
-                    log.error(" %s", e)
+                    log.estr(e)
     log("_load_dict_dir(%s)=%s", d, v)
     return v
 
-def _load_dict_dirs(dirname, parser) -> dict:
+def _load_dict_dirs(dirname:str, parser:Callable) -> Dict:
     if not GUESS_CONTENT:
         return {}
     #finds all the ".conf" files from the dirname specified
@@ -84,32 +86,33 @@ def _load_dict_dirs(dirname, parser) -> dict:
 # `content-type` mapping:
 ################################################################
 
-content_type_defs = None
-def load_content_type_defs() -> dict:
+content_type_defs : Optional[Dict] = None
+def load_content_type_defs() -> Dict:
     global content_type_defs
     if content_type_defs is None:
         content_type_defs = _load_dict_dirs("content-type", parse_content_types)
         if GUESS_CONTENT:
             #add env defs:
-            for entries in CONTENT_TYPE_DEFS.split(","):
-                content_type_defs.update(parse_content_types(entries))
+            content_type_defs.update(parse_content_types(CONTENT_TYPE_DEFS.split(",")))
     return content_type_defs
 
-def parse_content_types(lines) -> dict:
-    defs = {}
+def parse_content_types(lines) -> Dict[str,Dict[Any,Tuple[str,str]]]:
+    defs : Dict[str,Dict[Any,Tuple[str,str]]] = {}
     for line in lines:
+        if not line:
+            continue
         parts = line.rsplit("=", 1)
         #ie: "title:helloworld=text   #some comments here" -> "title:helloworld", "text   #some comments here"
         if len(parts)!=2:
             log.warn("Warning: invalid content-type definition")
-            log.warn(" %r is missing a '='", line)
+            log.warn(f" {line!r} is missing a '='")
             continue
         match_str, content_type = parts
         parts = match_str.split(":", 1)
         #ie: "title:helloworld" -> "title", "helloworld"
         if len(parts)!=2:
             log.warn("Warning: invalid content-type definition")
-            log.warn(" match string %r is missing a ':'", match_str)
+            log.warn(f" match string {match_str!r} is missing a ':'")
             continue
         #ignore comments:
         #"text    #some comments here" > "text"
@@ -119,8 +122,8 @@ def parse_content_types(lines) -> dict:
             c = re.compile(regex)
         except Exception as e:
             log.warn("Warning: invalid regular expression")
-            log.warn(" match string '%s':", regex)
-            log.warn(" %s", e)
+            log.warn(f" match string {regex!r}:")
+            log.warn(f" {e}")
             continue
         else:
             defs.setdefault(prop_name, {})[c]=(regex, content_type)
@@ -131,13 +134,12 @@ def get_content_type_properties():
     """ returns the list of window properties which can be used
         to guess the content-type.
     """
-    load_content_type_defs()
-    return content_type_defs.keys()
+    return load_content_type_defs().keys()
 
 
 def guess_content_type_from_defs(window) -> str:
-    global content_type_defs
     load_content_type_defs()
+    assert content_type_defs is not None
     for prop_name, defs in content_type_defs.items():
         if prop_name not in window.get_property_names():
             continue
@@ -156,13 +158,13 @@ def guess_content_type_from_defs(window) -> str:
                     log("guess_content_type(%s) found match: property=%s, regex=%s, content-type=%s",
                         window, prop_name, regex_str, content_type)
                     return content_type
-    return None
+    return ""
 
 ################################################################
 # `content-categories` mapping:
 ################################################################
 
-def parse_content_categories_file(lines) -> dict:
+def parse_content_categories_file(lines) -> Dict:
     d = {}
     for line in lines:
         parts = line.rsplit(":", 1)
@@ -176,7 +178,7 @@ def parse_content_categories_file(lines) -> dict:
     log("parse_content_categories_file(%s)=%s", lines, d)
     return d
 
-def load_categories_to_type() -> dict:
+def load_categories_to_type() -> Dict:
     return _load_dict_dirs("content-categories", parse_content_categories_file)
 
 
@@ -184,7 +186,7 @@ def load_categories_to_type() -> dict:
 # command mapping: using menu data
 ################################################################
 
-command_to_type = None
+command_to_type : Optional[Dict[str,str]] = None
 def load_command_to_type():
     global command_to_type
     if command_to_type is None:
@@ -213,7 +215,7 @@ def load_command_to_type():
                                         ctype = ct
                                         break
                             if ctype:
-                                cmd = os.path.basename(command.split(" ")[0]).encode()
+                                cmd = os.path.basename(command.split(" ")[0])
                                 if cmd:
                                     command_to_type[cmd] = ctype
                                     break
@@ -236,7 +238,7 @@ def guess_content_type_from_command(window):
 # `content-parent` mapping:
 ################################################################
 
-def parse_content_parent(lines):
+def parse_content_parent(lines) -> Dict[str,str]:
     v = {}
     for line in lines:
         parts = line.split(":", 1)
@@ -261,27 +263,28 @@ def guess_content_from_parent_pid(ppid):
     parent_command = get_proc_cmdline(ppid)
     if not parent_command:
         return None
-    try:
-        executable = os.path.basename(parent_command[0]).decode()
-    except UnicodeDecodeError:
-        executable = bytestostr(parent_command[0])
+    executable = os.path.basename(parent_command[0])
     pt = get_parent_to_type()
-    return pt.get(executable)
+    return pt.get(executable, "")
 
 
 def guess_content_type(window):
     if not GUESS_CONTENT:
         return DEFAULT_CONTENT_TYPE
-    return guess_content_type_from_defs(window) or guess_content_type_from_command(window) or guess_content_type_from_parent(window) or DEFAULT_CONTENT_TYPE
+    return guess_content_type_from_defs(window) or \
+        guess_content_type_from_command(window) or \
+        guess_content_type_from_parent(window) or \
+        DEFAULT_CONTENT_TYPE
 
 
 
 def main():
+    # pylint: disable=import-outside-toplevel
     import sys
     assert len(sys.argv)==2
     ppid = int(sys.argv[1])
     c = guess_content_from_parent_pid(ppid)
-    print("guess_content_from_parent_pid(%i)=%s" % (ppid, c))
+    print(f"guess_content_from_parent_pid({ppid})={c}")
 
 
 if __name__ == "__main__":  # pragma: no cover
