@@ -1,6 +1,6 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2011-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -8,16 +8,18 @@ import sys
 import os
 import socket
 import platform
+from typing import Any, Dict, Tuple, Optional
 
 #tricky: use xpra.scripts.config to get to the python "platform" module
 import xpra
-from xpra.util import updict, envbool, obsc, typedict, get_util_logger
-from xpra.os_util import get_linux_distribution, BITS, POSIX, WIN32
+from xpra.util import envbool, typedict, get_util_logger
+from xpra.os_util import get_linux_distribution, BITS, POSIX, WIN32, OSX
+from xpra.common import FULL_INFO
 
 XPRA_VERSION = xpra.__version__     #@UndefinedVariable
 
-CHECK_SSL = envbool("XPRA_VERSION_CHECK_SSL", True)
-SSL_CAFILE = None
+CHECK_SSL : bool = envbool("XPRA_VERSION_CHECK_SSL", True)
+SSL_CAFILE : str = ""
 if WIN32:
     try:
         import certifi  #@UnresolvedImport
@@ -27,16 +29,31 @@ if WIN32:
 SSL_CAFILE = os.environ.get("XPRA_SSL_CAFILE", SSL_CAFILE)
 
 
-def log(msg, *args, **kwargs):
+def log(msg, *args, **kwargs) -> None:
     get_util_logger().debug(msg, *args, **kwargs)
-def warn(msg, *args, **kwargs):
+def warn(msg, *args, **kwargs) -> None:
     get_util_logger().warn(msg, *args, **kwargs)
 
 
-def full_version_str() -> str:
+def vparts(vstr:str, n=1) -> str:
+    return ".".join(vstr.split(".")[:n])
+
+
+def version_str() -> str:
     rstr = revision_str()
     return XPRA_VERSION if not rstr else XPRA_VERSION+"-"+rstr
 
+def full_version_str() -> str:
+    rstr = version_str()
+    try:
+        from xpra.src_info import BRANCH  # pylint: disable=import-outside-toplevel
+    except ImportError:
+        pass
+    else:
+        if BRANCH=="master":
+            rstr += " beta"
+    return rstr
+
 def caps_to_version(caps : typedict) -> str:
     return caps.strget("version", "0")+"-"+caps_to_revision(caps)
 
@@ -59,92 +76,102 @@ def revision_str() -> str:
 def make_revision_str(revision, local_modifications, branch, commit) -> str:
     rstr = ""
     try:
-        rstr += "r"+revision
-        if local_modifications>0:
+        if isinstance(revision, int):
+            rstr += "r%i" % revision
+        if isinstance(local_modifications, int) and local_modifications>0:
             rstr += "M"
         if branch=="master" and commit:
             rstr += " (%s)" % commit
     except TypeError:
-        pass
+        get_util_logger().debug("make_revision_str%s", (revision, local_modifications, branch, commit), exc_info=True)
     return rstr
 
 
-
-def version_as_numbers(version : str):
-    return [int(x) for x in version.split(".")]
-
-def version_compat_check(remote_version : str):
+def version_compat_check(remote_version) -> Optional[str]:
     if remote_version is None:
         msg = "remote version not available!"
         log(msg)
         return msg
-    rv = version_as_numbers(remote_version)
-    lv = version_as_numbers(XPRA_VERSION)
+    try:
+        rv = parse_version(remote_version)
+    except ValueError:
+        warn(f"Warning: failed to parse remote version {remote_version!r}")
+        return None
+    try:
+        lv = parse_version(XPRA_VERSION)
+    except ValueError:
+        warn(f"Warning: failed to parse local version {XPRA_VERSION!r}")
+        return None
     if rv==lv:
         log("identical remote version: %s", remote_version)
         return None
-    if rv[0:3]<[0, 14, 10]:
-        #this is the oldest version we support
-        msg = "remote version %s is too old, sorry" % str(rv[:2])
+    try:
+        if rv[0:2] < (3, 0):
+            # this is the oldest version we support
+            msg = f"remote version {rv[:2]} is too old, sorry"
+            log(msg)
+            return msg
+    except TypeError as e:
+        msg = f"invalid remote version {rv[:2]}: {e}"
         log(msg)
         return msg
     if rv[0]>0:
-        log("newer remote version %s may work, we'll see..", remote_version)
+        log(f"newer remote version {remote_version} should work, we'll see..")
         return None
-    log("local version %s should be compatible with remote version: %s", XPRA_VERSION, remote_version)
+    log(f"local version {XPRA_VERSION!r} should be compatible with remote version {remote_version!r}")
     return None
 
 
-def get_host_info(obfuscate=False) -> dict:
+def get_host_info(full_info:int=1) -> Dict[str, Any]:
     #this function is for non UI thread info
-    info = {
-        "pid"                   : os.getpid(),
+    info : Dict[str, Any] = {}
+    if full_info>1:
+        info.update({
         "byteorder"             : sys.byteorder,
         "python"                : {
             "bits"                  : BITS,
             "full_version"          : sys.version,
             "version"               : ".".join(str(x) for x in sys.version_info[:3]),
             },
-        }
-    try:
-        hostname = socket.gethostname()
-        if obfuscate and hostname.find(".")>0:
-            parts = hostname.split(".")
-            for i, part in enumerate(parts):
-                if i>0:
-                    parts[i] = obsc(part)
-            hostname = ".".join(parts)
-        if hostname:
-            info["hostname"] = hostname
-    except OSError:
-        pass
-    if POSIX:
-        info.update({
-            "uid"   : os.getuid(),
-            "gid"   : os.getgid(),
-            })
+        })
+    if full_info>0:
+        try:
+            hostname = socket.gethostname()
+            if hostname:
+                info["hostname"] = hostname
+        except OSError:
+            pass
+        if POSIX:
+            info.update({
+                "uid"   : os.getuid(),
+                "gid"   : os.getgid(),
+                })
     return info
 
-def get_version_info() -> dict:
-    props = {
-             "version"  : XPRA_VERSION
-             }
-    try:
-        from xpra.src_info import LOCAL_MODIFICATIONS, REVISION, COMMIT, BRANCH
-        props.update({
-            "local_modifications"   : LOCAL_MODIFICATIONS,
-            "revision"              : REVISION,
-            "branch"                : BRANCH,
-            "commit"                : COMMIT,
-            })
-    except ImportError as e:
-        warn("missing some source information: %s", e)
-    return props
+def get_version_info(full:int=1) -> Dict[str, Any]:
+    info : Dict[str, Any] = {"version" : vparts(XPRA_VERSION, full+1)}
+    if full>0:
+        try:
+            # pylint: disable=import-outside-toplevel
+            from xpra.src_info import LOCAL_MODIFICATIONS, REVISION, COMMIT, BRANCH
+            for k,v in {
+                "local_modifications"   : LOCAL_MODIFICATIONS,
+                "revision"              : REVISION,
+                "branch"                : BRANCH,
+                "commit"                : COMMIT,
+                }.items():
+                if v is not None and v!="unknown":
+                    info[k] = v
+        except ImportError as e:
+            warn("missing some source information: %s", e)
+    if full>1:
+        info["build"] = get_build_info()
+    return info
 
-def get_version_info_full() -> dict:
-    props = get_version_info()
+def get_build_info(full:int=1) -> Dict[str, Any]:
+    info : Dict[str, Any] = {}
     try:
-        from xpra import build_info
+        from xpra import build_info  # pylint: disable=import-outside-toplevel
         #rename these build info properties:
         for k,bk in {
                     "date"                 : "BUILD_DATE",
@@ -158,17 +185,48 @@ def get_version_info_full() -> dict:
                     "cython"               : "CYTHON_VERSION",
                   }.items():
             v = getattr(build_info, bk, None)
-            if v:
-                props[k] = v
+            if v is not None:
+                if bk.endswith("_VERSION"):
+                    v = parse_version(v)
+                info[k] = v
         #record library versions:
-        d = dict((k.lstrip("lib_"), getattr(build_info, k)) for k in dir(build_info) if k.startswith("lib_"))
-        updict(props, "lib", d)
+        if full>1:
+            info["lib"] = dict((k.lstrip("lib_"), parse_version(getattr(build_info, k))) for k in dir(build_info) if k.startswith("lib_"))
     except Exception as e:
         warn("missing some build information: %s", e)
-    log("get_version_info_full()=%s", props)
-    return props
+    log(f"get_build_info({full})={info}")
+    return info
 
-def do_get_platform_info() -> dict:
+def parse_version(v) -> Tuple[Any, ...]:
+    if isinstance(v, str) and v:
+        def maybeint(v):
+            try:
+                return int(v)
+            except ValueError:
+                return v
+        v = tuple(maybeint(x) for x in v.split("-")[0].split("."))
+    return tuple(v or ())
+
+def vtrim(v, parts=FULL_INFO+1):
+    if isinstance(v, (list, tuple)):
+        return v[:parts]
+    return v
+
+def dict_version_trim(d, parts=FULL_INFO+1):
+    """
+    trims version numbers from info dictionaries
+    """
+    def vfilt(k, v):
+        if k.endswith("version") and isinstance(v, (list, tuple)):
+            v = vtrim(v, parts)
+        elif isinstance(v, dict):
+            return k, dict_version_trim(v, parts)
+        return k, v
+    return dict(vfilt(k,v) for k,v in d.items())
+
+
+def do_get_platform_info() -> Dict[str, Any]:
+    # pylint: disable=import-outside-toplevel
     from xpra.os_util import platform_name, platform_release
     pp = sys.modules.get("platform", platform)
     def get_processor_name():
@@ -177,20 +235,21 @@ def do_get_platform_info() -> dict:
         if pp.system() == "Darwin":
             os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
             command = ["sysctl", "-n", "machdep.cpu.brand_string"]
-            import subprocess
-            return subprocess.check_output(command).strip()
+            from subprocess import check_output
+            return check_output(command).strip()
         if pp.system() == "Linux":
-            with open("/proc/cpuinfo") as f:
+            with open("/proc/cpuinfo", "r", encoding="latin1") as f:
                 data = f.read()
             import re
             for line in data.split("\n"):
                 if "model name" in line:
-                    return re.sub(".*model name.*:", "", line,1).strip()
+                    return re.sub(".*model name.*:", "", line, count=1).strip()
         return pp.processor()
-    info = {}
-    ld = get_linux_distribution()
-    if ld:
-        info["linux_distribution"] = ld
+    info : Dict[str, Any] = {}
+    if POSIX and not OSX:
+        ld = get_linux_distribution()
+        if ld:
+            info["linux_distribution"] = ld
     try:
         release = platform_release(pp.release())
     except OSError:
@@ -216,11 +275,10 @@ def get_platform_info():
     return platform_info_cache
 
 
-def get_version_from_url(url):
-    e = None
+def get_version_from_url(url) -> Optional[Tuple[int,...]]:
     try:
         from urllib.request import urlopen
-    except ImportError:
+    except ImportError as e:
         log("get_version_from_url(%s) urllib2 not found: %s", url, e)
         return None
     try:
@@ -239,30 +297,46 @@ def get_version_from_url(url):
 
 def version_update_check():
     FAKE_NEW_VERSION = envbool("XPRA_FAKE_NEW_VERSION", False)
-    CURRENT_VERSION_URL = "%s://xpra.org/CURRENT_VERSION" % ("https" if CHECK_SSL else "http")
+    CURRENT_VERSION_URL = ("https" if CHECK_SSL else "http") + "://xpra.org/CURRENT_VERSION"
     PLATFORM_FRIENDLY_NAMES = {
         "linux2"    : "LINUX",
         "win"       : "WINDOWS",
         "darwin"    : "OSX",
         }
     our_version_no = tuple(int(y) for y in XPRA_VERSION.split("."))
+
+    BRANCH = os.environ.get("XPRA_CURRENT_VERSION_BRANCH", "")
+    if not BRANCH:
+        try:
+            from xpra.src_info import BRANCH
+        except ImportError as e:
+            log(f"unknown branch: {e}")
+    branch_strs = []
+    if BRANCH not in ("", "master"):
+        branch_parts = BRANCH.split(".")            # ie: "v6.2.x" -> ["v6", "2", "x"]
+        if branch_parts[-1] == "x":
+            branch_parts = branch_parts[:-1]        # ie: ["v6", "2"]
+        branch_strs.append("_" + "_".join(branch_parts[:2]))    # ie: "_v6_2"
+        if len(branch_parts) >= 2:
+            branch_strs.append("_" + branch_parts[0])           # ie: "_v6"
+    branch_strs.append("")
+
     platform_name = PLATFORM_FRIENDLY_NAMES.get(sys.platform, sys.platform)
     arch = get_platform_info().get("machine")
-    latest_version_no = None
-    for url in (
-        "%s_%s_%s?%s" % (CURRENT_VERSION_URL, platform_name, arch, XPRA_VERSION),
-        "%s_%s?%s" % (CURRENT_VERSION_URL, platform_name, XPRA_VERSION),
-        "%s?%s" % (CURRENT_VERSION_URL, XPRA_VERSION),
+    for branch_str in branch_strs:
+        for url in (
+            f"{CURRENT_VERSION_URL}{branch_str}_{platform_name}_{arch}?{XPRA_VERSION}",
+            f"{CURRENT_VERSION_URL}{branch_str}_{platform_name}?{XPRA_VERSION}",
+            f"{CURRENT_VERSION_URL}{branch_str}?{XPRA_VERSION}",
         ):
-        latest_version_no = get_version_from_url(url)
-        if latest_version_no:
-            break
+            latest_version_no = get_version_from_url(url)
+            if latest_version_no:
+                break
     if latest_version_no is None:
         log("version_update_check() failed to contact version server")
         return None
     if latest_version_no>our_version_no or FAKE_NEW_VERSION:
-        log("version_update_check() newer version found! local version is %s and the latest version available is %s",
-            our_version_no, latest_version_no)
-        #latest_version = ".".join([str(x) for x in latest_version_no])
+        log("version_update_check() newer version found")
+        log(" local version is {our_version_no} and the latest version available is {latest_version_no}")
         return latest_version_no
     return False
