@@ -1,17 +1,18 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+from typing import Dict, List, Tuple, Callable
 from gi.repository import GObject
 
-from xpra.util import WORKSPACE_UNSET, WORKSPACE_ALL
+from xpra.util import WORKSPACE_UNSET, WORKSPACE_ALL, first_time
 from xpra.x11.models.core import CoreX11WindowModel, xswallow, Above, RESTACKING_STR
-from xpra.x11.bindings.window_bindings import X11WindowBindings, constants      #@UnresolvedImport
+from xpra.x11.bindings.window import X11WindowBindings, constants      #@UnresolvedImport
 from xpra.server.window.content_guesser import guess_content_type, get_content_type_properties
-from xpra.x11.gtk_x11.gdk_bindings import get_pywindow, get_pyatom              #@UnresolvedImport
-from xpra.x11.gtk_x11.prop import prop_set, prop_get, prop_del
+from xpra.server.background_worker import add_work_item
+from xpra.x11.gtk3.gdk_bindings import get_pywindow, get_pyatom              #@UnresolvedImport
 from xpra.log import Logger
 
 log = Logger("x11", "window")
@@ -29,20 +30,20 @@ X11Window = X11WindowBindings()
 _NET_WM_STATE_REMOVE = 0
 _NET_WM_STATE_ADD    = 1
 _NET_WM_STATE_TOGGLE = 2
-STATE_STRING = {
+STATE_STRING : Dict[int,str] = {
     _NET_WM_STATE_REMOVE    : "REMOVE",
     _NET_WM_STATE_ADD       : "ADD",
     _NET_WM_STATE_TOGGLE    : "TOGGLE",
     }
 IconicState = constants["IconicState"]
 NormalState = constants["NormalState"]
-ICONIC_STATE_STRING = {
+ICONIC_STATE_STRING : Dict[int,str] = {
     IconicState  : "Iconic",
     NormalState  : "Normal",
     }
 
 #add user friendly workspace logging:
-WORKSPACE_STR = {
+WORKSPACE_STR : Dict[int,str] = {
     WORKSPACE_UNSET    : "UNSET",
     WORKSPACE_ALL      : "ALL",
     }
@@ -206,21 +207,21 @@ class BaseWindowModel(CoreX11WindowModel):
                               "_XPRA_SPEED",
                               "_XPRA_ENCODING",
                               ]
-    _DEFAULT_NET_WM_ALLOWED_ACTIONS = ["_NET_WM_ACTION_%s" % x for x in (
+    _DEFAULT_NET_WM_ALLOWED_ACTIONS : List[str] = [f"_NET_WM_ACTION_{x}" for x in (
         "CLOSE", "MOVE", "RESIZE", "FULLSCREEN",
         "MINIMIZE", "SHADE", "STICK",
         "MAXIMIZE_HORZ", "MAXIMIZE_VERT",
         "CHANGE_DESKTOP", "ABOVE", "BELOW")]
     _MODELTYPE = "Base"
 
-    def __init__(self, client_window):
-        super().__init__(client_window)
+    def __init__(self, xid:int):
+        super().__init__(xid)
         self.last_unmap_serial = 0
         self._input_field = True            # The WM_HINTS input field
         #watch for changes to properties that are used to derive the content-type:
         for x in get_content_type_properties():
             if x in self.get_dynamic_property_names():
-                self.connect("notify::%s" % x, self._content_type_related_property_change)
+                self.connect(f"notify::{x}", self._content_type_related_property_change)
 
     def serial_after_last_unmap(self, serial) -> bool:
         #"The serial member is set from the serial number reported in the protocol
@@ -233,17 +234,18 @@ class BaseWindowModel(CoreX11WindowModel):
         return False
 
 
-    def _read_initial_X11_properties(self):
+    def _read_initial_X11_properties(self) -> None:
         metalog("%s.read_initial_X11_properties()", self._MODELTYPE)
         self._updateprop("state", frozenset(self._read_wm_state()))
         super()._read_initial_X11_properties()
+        self._update_content_type()
 
     def _guess_window_type(self) -> str:
         #query the X11 property directly,
         #in case the python property isn't set yet
         if not self.is_OR():
             transient_for = self.prop_get("WM_TRANSIENT_FOR", "window")
-            if transient_for is not None:
+            if transient_for:
                 # EWMH says that even if it's transient-for, we MUST check to
                 # see if it's override-redirect (and if so treat as NORMAL).
                 # But we wouldn't be here if this was override-redirect.
@@ -256,7 +258,7 @@ class BaseWindowModel(CoreX11WindowModel):
     # Actions
     ################################
 
-    def move_to_workspace(self, workspace : int):
+    def move_to_workspace(self, workspace : int) -> None:
         #we send a message to ourselves, we could also just update the property
         current = self.get_property("workspace")
         if current==workspace:
@@ -266,27 +268,27 @@ class BaseWindowModel(CoreX11WindowModel):
         with xswallow:
             if workspace==WORKSPACE_UNSET:
                 workspacelog("removing _NET_WM_DESKTOP property from window %#x", self.xid)
-                prop_del(self.client_window, "_NET_WM_DESKTOP")
+                self.prop_del("_NET_WM_DESKTOP")
             else:
                 workspacelog("setting _NET_WM_DESKTOP=%s on window %#x", workspacestr(workspace), self.xid)
-                prop_set(self.client_window, "_NET_WM_DESKTOP", "u32", workspace)
+                self.prop_set("_NET_WM_DESKTOP", "u32", workspace)
 
 
     #########################################
     # Python objects synced to X11 properties
     #########################################
 
-    def _sync_state(self, *_args):
+    def _sync_state(self, *_args) -> None:
         state = self.get_property("state")
         metalog("sync_state: setting _NET_WM_STATE=%s on %#x", state, self.xid)
         with xswallow:
-            prop_set(self.client_window, "_NET_WM_STATE", ["atom"], state)
+            self.prop_set("_NET_WM_STATE", ["atom"], state)
 
-    def _sync_iconic(self, *_args):
+    def _sync_iconic(self, *_args) -> None:
         def set_state(state):
             log("_handle_iconic_update: set_state(%s)", state)
             with xswallow:
-                prop_set(self.client_window, "WM_STATE", "state", state)
+                self.prop_set("WM_STATE", "state", state)
 
         if self.get("iconic"):
             set_state(IconicState)
@@ -296,7 +298,7 @@ class BaseWindowModel(CoreX11WindowModel):
             self._state_remove("_NET_WM_STATE_HIDDEN")
 
 
-    _py_property_handlers = dict(CoreX11WindowModel._py_property_handlers)
+    _py_property_handlers : Dict[str,Callable] = dict(CoreX11WindowModel._py_property_handlers)
     _py_property_handlers.update({
         "state"         : _sync_state,
         "iconic"        : _sync_iconic,
@@ -307,13 +309,14 @@ class BaseWindowModel(CoreX11WindowModel):
     # X11 properties synced to Python objects
     #########################################
 
-    def _handle_transient_for_change(self):
+    def _handle_transient_for_change(self) -> None:
         transient_for = self.prop_get("WM_TRANSIENT_FOR", "window")
         metalog("WM_TRANSIENT_FOR=%s", transient_for)
         # May be None
-        self._updateprop("transient-for", transient_for)
+        window = get_pywindow(transient_for) if transient_for else None
+        self._updateprop("transient-for", window)
 
-    def _handle_window_type_change(self):
+    def _handle_window_type_change(self) -> None:
         window_types = self.prop_get("_NET_WM_WINDOW_TYPE", ["atom"])
         metalog("_NET_WM_WINDOW_TYPE=%s", window_types)
         if not window_types:
@@ -325,24 +328,24 @@ class BaseWindowModel(CoreX11WindowModel):
         window_types = [str(wt).replace("_NET_WM_WINDOW_TYPE_", "").replace("_NET_WM_TYPE_", "") for wt in window_types]
         self._updateprop("window-type", window_types)
 
-    def _handle_workspace_change(self):
+    def _handle_workspace_change(self) -> None:
         workspace = self.prop_get("_NET_WM_DESKTOP", "u32", True)
         if workspace is None:
             workspace = WORKSPACE_UNSET
         workspacelog("_NET_WM_DESKTOP=%s for window %#x", workspacestr(workspace), self.xid)
         self._updateprop("workspace", workspace)
 
-    def _handle_fullscreen_monitors_change(self):
+    def _handle_fullscreen_monitors_change(self) -> None:
         fsm = self.prop_get("_NET_WM_FULLSCREEN_MONITORS", ["u32"], True)
         metalog("_NET_WM_FULLSCREEN_MONITORS=%s", fsm)
         self._updateprop("fullscreen-monitors", fsm)
 
-    def _handle_bypass_compositor_change(self):
+    def _handle_bypass_compositor_change(self) -> None:
         bypass = self.prop_get("_NET_WM_BYPASS_COMPOSITOR", "u32", True) or 0
         metalog("_NET_WM_BYPASS_COMPOSITOR=%s", bypass)
         self._updateprop("bypass-compositor", bypass)
 
-    def _handle_wm_strut_change(self):
+    def _handle_wm_strut_change(self) -> None:
         strut = self.prop_get("_NET_WM_STRUT_PARTIAL", "strut-partial")
         metalog("_NET_WM_STRUT_PARTIAL=%s", strut)
         if strut is None:
@@ -351,12 +354,12 @@ class BaseWindowModel(CoreX11WindowModel):
         # Might be None:
         self._updateprop("strut", strut)
 
-    def _handle_opacity_change(self):
+    def _handle_opacity_change(self) -> None:
         opacity = self.prop_get("_NET_WM_WINDOW_OPACITY", "u32", True) or -1
         metalog("_NET_WM_WINDOW_OPACITY=%s", opacity)
         self._updateprop("opacity", opacity)
 
-    def _handle_wm_hints_change(self):
+    def _handle_wm_hints_change(self) -> None:
         with xswallow:
             wm_hints = X11Window.getWMHints(self.xid)
         metalog("getWMHints(%#x)=%s", self.xid, wm_hints)
@@ -367,10 +370,9 @@ class BaseWindowModel(CoreX11WindowModel):
         if "window_group" in wm_hints:
             xid = wm_hints.get("window_group")
             if xid:
-                try:
-                    group_leader = get_pywindow(xid)
-                except Exception:
-                    log.error("Error locating group leader window %#x", xid)
+                group_leader = get_pywindow(xid)
+                if not group_leader:
+                    log.error(f"Error group leader window {xid:x} not found")
         self._updateprop("group-leader", group_leader)
         self._updateprop("attention-requested", wm_hints.get("urgency", False))
         _input = wm_hints.get("input")
@@ -382,46 +384,54 @@ class BaseWindowModel(CoreX11WindowModel):
             self._input_field = int(_input)
             self._update_can_focus()
 
-    def _update_can_focus(self, *_args):
+    def _update_can_focus(self, *_args) -> None:
         can_focus = bool(self._input_field) or "WM_TAKE_FOCUS" in self.get_property("protocols")
         self._updateprop("can-focus", can_focus)
 
 
-    def _content_type_related_property_change(self, *_args):
+    def _content_type_related_property_change(self, *_args) -> None:
         self._update_content_type()
 
-    def _handle_xpra_content_type_change(self):
+    def _handle_xpra_content_type_change(self) -> None:
         self._update_content_type()
 
-    def _update_content_type(self):
-        #watch for changes to properties that are used to derive the content-type:
-        content_type = self.prop_get("_XPRA_CONTENT_TYPE", "latin1", True)
-        #the _XPRA_CONTENT_TYPE property takes precedence
+    def _update_content_type(self) -> None:
+        # watch for changes to properties that are used to derive the content-type:
+        content_type = self.prop_get("_XPRA_CONTENT_TYPE", "latin1", True) or ""
+        # the _XPRA_CONTENT_TYPE property takes precedence
         if not content_type:
-            content_type = guess_content_type(self)
-            if not content_type and self.is_tray():
-                content_type = "picture"
+            # guess_content_type can take a while as it will load the xdg menu data
+            # so call it as a work item.
+            # multiple callers can be added to the sequential work queue,
+            # the menu loader also has a lock.
+            add_work_item(self.guess_content_type)
         metalog("_update_content_type() %s", content_type)
         self._updateprop("content-type", content_type)
 
+    def guess_content_type(self) -> None:
+        content_type = guess_content_type(self)
+        if not content_type and self.is_tray():
+            content_type = "picture"
+        metalog("guess_content_type() content_type=%s", content_type)
+        self._updateprop("content-type", content_type)
 
-    def _handle_xpra_quality_change(self):
+    def _handle_xpra_quality_change(self) -> None:
         quality = self.prop_get("_XPRA_QUALITY", "u32", True) or -1
         metalog("quality=%s", quality)
         self._updateprop("quality", max(-1, min(100, quality)))
 
-    def _handle_xpra_speed_change(self):
+    def _handle_xpra_speed_change(self) -> None:
         speed = self.prop_get("_XPRA_SPEED", "u32", True) or -1
         metalog("speed=%s", speed)
         self._updateprop("speed", max(-1, min(100, speed)))
 
-    def _handle_xpra_encoding_change(self):
+    def _handle_xpra_encoding_change(self) -> None:
         encoding = self.prop_get("_XPRA_ENCODING", "latin1", True) or ""
         metalog("encoding=%s", encoding)
         self._updateprop("encoding", encoding)
 
 
-    _x11_property_handlers = CoreX11WindowModel._x11_property_handlers.copy()
+    _x11_property_handlers : Dict[str,Callable] = CoreX11WindowModel._x11_property_handlers.copy()
     _x11_property_handlers.update({
         "WM_TRANSIENT_FOR"              : _handle_transient_for_change,
         "_NET_WM_WINDOW_TYPE"           : _handle_window_type_change,
@@ -458,7 +468,7 @@ class BaseWindowModel(CoreX11WindowModel):
     #      directly by the "state" property, and reading/writing them in fact
     #      accesses the "state" set directly.  This is done by overriding
     #      do_set_property and do_get_property.
-    _state_properties = {
+    _state_properties : Dict[str,Tuple[str,...]] = {
         "attention-requested"   : ("_NET_WM_STATE_DEMANDS_ATTENTION", ),
         "fullscreen"            : ("_NET_WM_STATE_FULLSCREEN", ),
         "maximized"             : ("_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ"),
@@ -476,7 +486,7 @@ class BaseWindowModel(CoreX11WindowModel):
         for x in states:
             _state_properties_reversed[x] = k
 
-    def _state_add(self, *state_names):
+    def _state_add(self, *state_names) -> None:
         curr = set(self.get_property("state"))
         add = [s for s in state_names if s not in curr]
         if add:
@@ -486,7 +496,7 @@ class BaseWindowModel(CoreX11WindowModel):
             self._internal_set_property("state", frozenset(curr))
             self._state_notify(add)
 
-    def _state_remove(self, *state_names):
+    def _state_remove(self, *state_names) -> None:
         curr = set(self.get_property("state"))
         discard = [s for s in state_names if s in curr]
         if discard:
@@ -496,7 +506,7 @@ class BaseWindowModel(CoreX11WindowModel):
             self._internal_set_property("state", frozenset(curr))
             self._state_notify(discard)
 
-    def _state_notify(self, state_names):
+    def _state_notify(self, state_names) -> None:
         notify_props = set()
         for x in state_names:
             if x in self._state_properties_reversed:
@@ -504,16 +514,16 @@ class BaseWindowModel(CoreX11WindowModel):
         for x in tuple(notify_props):
             self.notify(x)
 
-    def _state_isset(self, state_name):
+    def _state_isset(self, state_name) -> bool:
         return state_name in self.get_property("state")
 
-    def _read_wm_state(self):
+    def _read_wm_state(self) -> List[str]:
         wm_state = self.prop_get("_NET_WM_STATE", ["atom"])
         metalog("read _NET_WM_STATE=%s", wm_state)
         return wm_state or []
 
 
-    def do_set_property(self, pspec, value):
+    def do_set_property(self, pspec, value) -> None:
         #intercept state properties to route via update_state()
         if pspec.name in self._state_properties:
             #virtual property for WM_STATE:
@@ -521,7 +531,7 @@ class BaseWindowModel(CoreX11WindowModel):
             return
         super().do_set_property(pspec, value)
 
-    def do_get_property(self, pspec):
+    def do_get_property(self, pspec) -> object:
         #intercept state properties to route via get_wm_state()
         if pspec.name in self._state_properties:
             #virtual property for WM_STATE:
@@ -529,17 +539,19 @@ class BaseWindowModel(CoreX11WindowModel):
         return super().do_get_property(pspec)
 
 
-    def update_wm_state(self, prop, b):
+    def update_wm_state(self, prop, b) -> None:
         state_names = self._state_properties.get(prop)
-        assert state_names, "invalid window state %s" % prop
+        if not state_names:
+            raise ValueError(f"invalid window state {prop}")
         if b:
             self._state_add(*state_names)
         else:
             self._state_remove(*state_names)
 
-    def get_wm_state(self, prop):
+    def get_wm_state(self, prop) -> bool:
         state_names = self._state_properties.get(prop)
-        assert state_names, "invalid window state %s" % prop
+        if not state_names:
+            raise ValueError(f"invalid window state {prop}")
         #this is a virtual property for WM_STATE:
         #return True if any is set (only relevant for maximized)
         for x in state_names:
@@ -552,7 +564,7 @@ class BaseWindowModel(CoreX11WindowModel):
     # X11 Events
     #########################################
 
-    def process_client_message_event(self, event):
+    def process_client_message_event(self, event) -> bool:
         # FIXME
         # Need to listen for:
         #   _NET_CURRENT_DESKTOP
@@ -571,7 +583,7 @@ class BaseWindowModel(CoreX11WindowModel):
                 elif mode==_NET_WM_STATE_TOGGLE:
                     v = not bool(current)
                 else:
-                    log.warn("Warning: invalid mode for _NET_WM_STATE: %s", mode)
+                    log.warn(f"Warning: invalid mode for _NET_WM_STATE: {mode}")
                     return
                 log("process_client_message_event(%s) window %s=%s after %s (current state=%s)",
                     event, prop, v, STATE_STRING.get(mode, mode), current)
@@ -607,9 +619,12 @@ class BaseWindowModel(CoreX11WindowModel):
                 update_wm_state("modal")
             elif atom1=="_NET_WM_STATE_DEMANDS_ATTENTION":
                 update_wm_state("attention-requested")
+            elif atom1 == "_KDE_NET_WM_STATE_SKIP_SWITCHER":
+                if first_time(f"KDE-WM_STATE-{self.xid}"):
+                    log.info(f"invalid KDE specific request {atom1} ignored, will not be logged again")
             else:
-                log.info("Unhandled _NET_WM_STATE request: '%s'", event, atom1)
-                log.info(" event%s", event)
+                log.info(f"Unhandled _NET_WM_STATE request: {atom1!r}")
+                log.info(f" event {event!r}")
             return True
         if event.message_type=="WM_CHANGE_STATE":
             iconic = event.data[0]
@@ -632,14 +647,13 @@ class BaseWindowModel(CoreX11WindowModel):
             #if event.data[0] in ACTIVE_WINDOW_SOURCE:
             log("_NET_ACTIVE_WINDOW: %s", event)
             self.set_active()
-            self.emit("restack", Above, None)
+            self.emit("restack", Above, 0)
             return True
         if event.message_type=="_NET_WM_DESKTOP":
             workspace = int(event.data[0])
             #query the workspace count on the root window
             #since we cannot access Wm from here..
-            root = self.client_window.get_screen().get_root_window()
-            ndesktops = prop_get(root, "_NET_NUMBER_OF_DESKTOPS", "u32", ignore_errors=True)
+            ndesktops = int(self.root_prop_get("_NET_NUMBER_OF_DESKTOPS", "u32") or 0)
             workspacelog("received _NET_WM_DESKTOP: workspace=%s, number of desktops=%s",
                          workspacestr(workspace), ndesktops)
             if ndesktops>0 and (
@@ -661,13 +675,14 @@ class BaseWindowModel(CoreX11WindowModel):
                 log.warn("Warning: invalid list of _NET_WM_FULLSCREEN_MONITORS:%s - ignored", event.data)
                 return False
             log("_NET_WM_FULLSCREEN_MONITORS: monitors=%s", monitors)
-            prop_set(self.client_window, "_NET_WM_FULLSCREEN_MONITORS", ["u32"], monitors)
+            self.prop_set("_NET_WM_FULLSCREEN_MONITORS", ["u32"], monitors)
             return True
         if event.message_type=="_NET_RESTACK_WINDOW":
             source = {1 : "application", 2 : "pager"}.get(event.data[0], "default (%s)" % event.data[0])
             sibling_window = event.data[1]
             log("%s sent to window %#x for sibling %#x from %s with detail=%s",
-                event.message_type, event.window, sibling_window, source, RESTACKING_STR.get(event.detail, event.detail))
+                event.message_type, event.window, sibling_window, source,
+                RESTACKING_STR.get(event.detail, event.detail))
             self.emit("restack", event.detail, sibling_window)
             return True
         #TODO: maybe we should process _NET_MOVERESIZE_WINDOW here?
