@@ -1,32 +1,27 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2022 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-from gi.repository import GObject, Gtk
+from typing import Tuple
+
+from gi.repository import GObject   # @UnresolvedImport
 
 from xpra.util import envint, envbool, typedict
 from xpra.common import MAX_WINDOW_SIZE
 from xpra.gtk_common.gobject_util import one_arg_signal
 from xpra.gtk_common.error import XError, xsync, xswallow, xlog
-from xpra.x11.gtk_x11.send_wm import send_wm_take_focus
-from xpra.x11.gtk_x11.prop import prop_set, prop_get
+from xpra.x11.gtk_x11.prop import prop_set
 from xpra.x11.prop_conv import MotifWMHints
-from xpra.x11.bindings.window_bindings import X11WindowBindings #@UnresolvedImport
+from xpra.x11.bindings.window import X11WindowBindings #@UnresolvedImport
+from xpra.x11.bindings.send_wm import send_wm_take_focus
 from xpra.x11.common import Unmanageable
 from xpra.x11.models.size_hints_util import sanitize_size_hints
 from xpra.x11.models.base import BaseWindowModel, constants
 from xpra.x11.models.core import sanestr
-from xpra.x11.gtk_x11.gdk_bindings import (
-    add_event_receiver, remove_event_receiver,
-    get_children,
-    calc_constrained_size,
-    x11_get_server_time,
-    get_pywindow,
-    )
-
-from xpra.gtk_common.gtk_util import get_default_root_window
+from xpra.x11.gtk3.gdk_bindings import add_event_receiver, remove_event_receiver
+from xpra.gtk_common.gtk3.gdk_bindings import calc_constrained_size
 from xpra.log import Logger
 
 log = Logger("x11", "window")
@@ -57,7 +52,6 @@ CW_MASK_TO_NAME = {
                    CWY              : "Y",
                    CWWidth          : "Width",
                    CWHeight         : "Height",
-                   CWBorderWidth    : "BorderWidth",
                    CWSibling        : "Sibling",
                    CWStackMode      : "StackMode",
                    CWBorderWidth    : "BorderWidth",
@@ -66,7 +60,7 @@ def configure_bits(value_mask):
     return "|".join(v for k,v in CW_MASK_TO_NAME.items() if k&value_mask)
 
 
-FORCE_XSETINPUTFOCUS = envbool("XPRA_FORCE_XSETINPUTFOCUS", True)
+FORCE_XSETINPUTFOCUS = envbool("XPRA_FORCE_XSETINPUTFOCUS", False)
 VALIDATE_CONFIGURE_REQUEST = envbool("XPRA_VALIDATE_CONFIGURE_REQUEST", False)
 CLAMP_OVERLAP = envint("XPRA_WINDOW_CLAMP_OVERLAP", 20)
 assert CLAMP_OVERLAP>=0
@@ -76,7 +70,7 @@ class WindowModel(BaseWindowModel):
     """This represents a managed client window.  It allows one to produce
     widgets that view that client window in various ways."""
 
-    _NET_WM_ALLOWED_ACTIONS = ["_NET_WM_ACTION_%s" % x for x in (
+    _NET_WM_ALLOWED_ACTIONS = [f"_NET_WM_ACTION_{x}" for x in (
         "CLOSE", "MOVE", "RESIZE", "FULLSCREEN",
         "MINIMIZE", "SHADE", "STICK",
         "MAXIMIZE_HORZ", "MAXIMIZE_VERT",
@@ -112,10 +106,10 @@ class WindowModel(BaseWindowModel):
                                  "Should the requested position be honoured?", "",
                                  False,
                                  GObject.ParamFlags.READWRITE),
-        # Toggling this property does not actually make the window iconified,
+        # Toggling this property does not actually make the window `iconified`,
         # i.e. make it appear or disappear from the screen -- it merely
         # updates the various window manager properties that inform the world
-        # whether or not the window is iconified.
+        # whether or not the window is `iconified`.
         "iconic": (GObject.TYPE_BOOLEAN,
                    "ICCCM 'iconic' state -- any sort of 'not on desktop'.", "",
                    False,
@@ -162,22 +156,23 @@ class WindowModel(BaseWindowModel):
     _internal_property_names = BaseWindowModel._internal_property_names+["children"]
     _MODELTYPE = "Window"
 
-    def __init__(self, parking_window, client_window, desktop_geometry, size_constraints=None):
+    def __init__(self, parking_window_xid:int, xid:int, desktop_geometry, size_constraints=None):
         """Register a new client window with the WM.
 
         Raises an Unmanageable exception if this window should not be
         managed, for whatever reason.  ATM, this mostly means that the window
         died somehow before we could do anything with it."""
 
-        super().__init__(client_window)
-        self.parking_window = parking_window
-        self.corral_window = None
+        super().__init__(xid)
+        self.parking_window_xid = parking_window_xid
+        self.corral_xid : int = 0
         self.desktop_geometry = desktop_geometry
         self.size_constraints = size_constraints or (0, 0, MAX_WINDOW_SIZE, MAX_WINDOW_SIZE)
-        #extra state attributes so we can unmanage() the window cleanly:
-        self.in_save_set = False
-        self.client_reparented = False
-        self.kill_count = 0
+        self.saved_events = -1
+        # these extra state attributes allow us to `unmanage()` the window cleanly:
+        self.in_save_set : bool = False
+        self.client_reparented : bool = False
+        self.kill_count : int = 0
 
         self.call_setup()
 
@@ -185,23 +180,19 @@ class WindowModel(BaseWindowModel):
     # Setup and teardown
     #########################################
 
-    def setup(self):
+    def setup(self) -> None:
         super().setup()
-
-        ogeom = self.client_window.get_geometry()
+        self.saved_events = X11Window.getEventMask(self.xid)
+        ogeom = X11Window.getGeometry(self.xid)
         ox, oy, ow, oh = ogeom[:4]
-        # We enable PROPERTY_CHANGE_MASK so that we can call
-        # x11_get_server_time on this window.
         # clamp this window to the desktop size:
         x, y = self._clamp_to_desktop(ox, oy, ow, oh)
         geomlog("setup() clamp_to_desktop(%s)=%s", ogeom, (x, y))
-        parking_xid = self.parking_window.get_xid()
-        cxid = X11Window.CreateCorralWindow(parking_xid, self.xid, x, y)
-        self.corral_window = get_pywindow(cxid)
-        log("setup() corral_xid=%#x, corral_window=", cxid, self.corral_window)
-        prop_set(self.corral_window, "_NET_WM_NAME", "utf8", "Xpra-CorralWindow-%#x" % self.xid)
-        X11Window.substructureRedirect(cxid)
-        add_event_receiver(self.corral_window, self)
+        self.corral_xid = X11Window.CreateCorralWindow(self.parking_window_xid, self.xid, x, y)
+        log("setup() corral_xid=%#x", self.corral_xid)
+        prop_set(self.corral_xid, "_NET_WM_NAME", "utf8", "Xpra-CorralWindow-%#x" % self.xid)
+        X11Window.substructureRedirect(self.corral_xid)
+        add_event_receiver(self.corral_xid, self)
 
         # The child might already be mapped, in case we inherited it from
         # a previous window manager.  If so, we unmap it now, and save the
@@ -217,12 +208,12 @@ class WindowModel(BaseWindowModel):
         self.in_save_set = True
 
         log("setup() reparenting")
-        X11Window.Reparent(self.xid, cxid, 0, 0)
+        X11Window.Reparent(self.xid, self.corral_xid, 0, 0)
         self.client_reparented = True
 
         geom = X11Window.geometry_with_border(self.xid)
         if geom is None:
-            raise Unmanageable("window %#x disappeared already" % self.xid)
+            raise Unmanageable(f"window {self.xid:x} disappeared already")
         geomlog("setup() geometry=%s, ogeom=%s", geom, ogeom)
         nx, ny, w, h = geom[:4]
         #after reparenting, the coordinates of the client window should be 0,0
@@ -242,24 +233,21 @@ class WindowModel(BaseWindowModel):
         self._updateprop("geometry", (nx, ny, nw, nh))
         geomlog("setup() resizing windows to %sx%s, moving to %i,%i", nw, nh, nx, ny)
         #don't trigger a move or resize unless we have to:
-        if (ox,oy)!=(nx,ny) and (ow,oh)!=(nw,nh):
-            self.corral_window.move_resize(nx, ny, nw, nh)
-        elif (ox,oy)!=(nx,ny):
-            self.corral_window.move(nx, ny)
-        elif (ow,oh)!=(nw,nh):
-            self.corral_window.resize(nw, nh)
-        if (ow,oh)!=(nw,nh):
-            self.client_window.resize(nw, nh)
-        self.client_window.show_unraised()
+        if ox!=nx or oy!=ny or ow!=nw or oh!=nh:
+            X11Window.MoveResizeWindow(self.corral_xid, nx, ny, nw, nh)
+        if ow!=nw or oh!=nh:
+            X11Window.MoveResizeWindow(self.xid, 0, 0, nw, nh)
+        if not X11Window.is_mapped(self.xid):
+            X11Window.MapWindow(self.xid)
         #this is here to trigger X11 errors if any are pending
         #or if the window is deleted already:
-        self.client_window.get_geometry()
+        assert X11Window.getGeometry(self.xid)
         self._internal_set_property("shown", False)
         self._internal_set_property("resize-counter", 0)
         self._internal_set_property("client-geometry", None)
 
 
-    def _clamp_to_desktop(self, x, y, w, h):
+    def _clamp_to_desktop(self, x:int, y:int, w:int, h:int) -> Tuple[int,int]:
         if self.desktop_geometry:
             dw, dh = self.desktop_geometry
             if x+w<0:
@@ -272,18 +260,19 @@ class WindowModel(BaseWindowModel):
                 y = max(0, dh-CLAMP_OVERLAP)
         return x, y
 
-    def update_desktop_geometry(self, width, height):
+    def update_desktop_geometry(self, width:int, height:int) -> None:
         if self.desktop_geometry==(width, height):
             return  #no need to do anything
         self.desktop_geometry = (width, height)
-        x, y, w, h = self.corral_window.get_geometry()[:4]
-        nx, ny = self._clamp_to_desktop(x, y, w, h)
-        if nx!=x or ny!=y:
-            log("update_desktop_geometry(%i, %i) adjusting corral window to new location: %i,%i", width, height, nx, ny)
-            self.corral_window.move(nx, ny)
+        with xsync:
+            x, y, w, h = X11Window.getGeometry(self.corral_xid)[:4]
+            nx, ny = self._clamp_to_desktop(x, y, w, h)
+            if nx!=x or ny!=y:
+                log("update_desktop_geometry(%i, %i) adjusting corral window to new location: %i,%i", width, height, nx, ny)
+                X11Window.MoveResizeWindow(self.corral_xid, nx, ny, w, h)
 
 
-    def _read_initial_X11_properties(self):
+    def _read_initial_X11_properties(self) -> None:
         metalog("read_initial_X11_properties() window")
         # WARNING: have to handle _NET_WM_STATE before we look at WM_HINTS;
         # WM_HINTS assumes that our "state" property is already set.  This is
@@ -317,7 +306,7 @@ class WindowModel(BaseWindowModel):
         assert net_wm_state is not None, "_NET_WM_STATE should have been read already"
         geom = X11Window.getGeometry(self.xid)
         if not geom:
-            raise Unmanageable("failed to get geometry for %#x" % self.xid)
+            raise Unmanageable(f"failed to get geometry for {self.xid:x}")
         #initial position and size, from the Window object,
         #but allow size hints to override it if specified
         x, y, w, h = geom[:4]
@@ -341,21 +330,24 @@ class WindowModel(BaseWindowModel):
             self._internal_set_property("set-initial-position", False)
         self.update_children()
 
-    def do_unmanaged(self, wm_exiting):
-        log("unmanaging window: %s (%s - %s)", self, self.corral_window, self.client_window)
-        cwin = self.corral_window
-        if cwin:
-            self.corral_window = None
-            remove_event_receiver(cwin, self)
+    def do_unmanaged(self, wm_exiting) -> None:
+        log("unmanaging window: %s (%s - %s)", self, self.corral_xid, self.xid)
+        cxid = self.corral_xid
+        if cxid:
+            self.corral_xid = 0
+            remove_event_receiver(cxid, self)
             geom = None
-            #use a new context so we will XSync right here
-            #and detect if the window is already gone:
+            # use a new context, so we will XSync right here
+            # and detect if the window is already gone:
             with xswallow:
                 geom = X11Window.getGeometry(self.xid)
             if geom is not None:
-                if self.client_reparented:
-                    self.client_window.reparent(get_default_root_window(), 0, 0)
-                self.client_window.set_events(self.client_window_saved_events)
+                with xswallow:
+                    if self.client_reparented:
+                        X11Window.Reparent(self.xid, X11Window.get_root_xid(), 0, 0)
+                    if self.saved_events!=-1:
+                        with xswallow:
+                            X11Window.setEventMask(self.xid, self.saved_events)
             self.client_reparented = False
             # It is important to remove from our save set, even after
             # reparenting, because according to the X spec, windows that are
@@ -367,34 +359,38 @@ class WindowModel(BaseWindowModel):
                 with xswallow:
                     X11Window.XRemoveFromSaveSet(self.xid)
                 self.in_save_set = False
-            with xswallow:
-                X11Window.sendConfigureNotify(self.xid)
+            if geom:
+                self.send_configure_notify()
             if wm_exiting:
-                self.client_window.show_unraised()
+                with xswallow:
+                    if not X11Window.is_mapped(self.xid):
+                        X11Window.MapWindow(self.xid)
             #it is now safe to destroy the corral window:
-            cxid = cwin.get_xid()
-            if cxid:
-                with xsync:
-                    X11Window.DestroyWindow(cxid)
+            with xsync:
+                X11Window.DestroyWindow(cxid)
         super().do_unmanaged(wm_exiting)
 
+    def send_configure_notify(self) -> None:
+        with xswallow:
+            X11Window.sendConfigureNotify(self.xid)
+
 
     #########################################
     # Actions specific to WindowModel
     #########################################
 
-    def raise_window(self):
-        X11Window.XRaiseWindow(self.corral_window.get_xid())
-        X11Window.XRaiseWindow(self.client_window.get_xid())
+    def raise_window(self) -> None:
+        X11Window.XRaiseWindow(self.corral_xid)
+        X11Window.XRaiseWindow(self.xid)
 
-    def unmap(self):
+    def unmap(self) -> None:
         with xsync:
             if X11Window.is_mapped(self.xid):
                 self.last_unmap_serial = X11Window.Unmap(self.xid)
                 log("client window %#x unmapped, serial=%#x", self.xid, self.last_unmap_serial)
                 self._internal_set_property("shown", False)
 
-    def map(self):
+    def map(self) -> None:
         with xsync:
             if not X11Window.is_mapped(self.xid):
                 X11Window.MapWindow(self.xid)
@@ -405,12 +401,12 @@ class WindowModel(BaseWindowModel):
     # X11 Events
     #########################################
 
-    def do_xpra_property_notify_event(self, event):
-        if event.delivered_to is self.corral_window:
+    def do_xpra_property_notify_event(self, event) -> None:
+        if event.delivered_to==self.corral_xid:
             return
         super().do_xpra_property_notify_event(event)
 
-    def do_child_map_request_event(self, event):
+    def do_child_map_request_event(self, event) -> None:
         # If we get a MapRequest then it might mean that someone tried to map
         # this window multiple times in quick succession, before we actually
         # mapped it (so that several MapRequests ended up queued up; FSF Emacs
@@ -420,13 +416,14 @@ class WindowModel(BaseWindowModel):
         # request.
         log("do_child_map_request_event(%s)", event)
 
-    def do_xpra_unmap_event(self, event):
-        if event.delivered_to is self.corral_window or self.corral_window is None:
+    def do_xpra_unmap_event(self, event) -> None:
+        log(f"do_xpra_unmap_event({event}) corral_xid={self.corral_xid:x}")
+        if not self.corral_xid or event.delivered_to==self.corral_xid:
             return
-        assert event.window is self.client_window
+        assert event.window==self.xid
         # The client window got unmapped.  The question is, though, was that
         # because it was withdrawn/destroyed, or was it because we unmapped it
-        # going into IconicState?
+        # going into `IconicState`?
         #
         # Also, if we receive a *synthetic* UnmapNotify event, that always
         # means that the client has withdrawn the window (even if it was not
@@ -435,10 +432,11 @@ class WindowModel(BaseWindowModel):
         if event.send_event or self.serial_after_last_unmap(event.serial):
             self.unmanage()
 
-    def do_xpra_destroy_event(self, event):
-        if event.delivered_to is self.corral_window or self.corral_window is None:
+    def do_xpra_destroy_event(self, event) -> None:
+        log(f"do_xpra_destroy_event({event}) corral_xid={self.corral_xid:x}")
+        if not self.corral_xid or event.delivered_to==self.corral_xid:
             return
-        assert event.window is self.client_window
+        assert event.window==self.xid
         super().do_xpra_destroy_event(event)
 
 
@@ -446,32 +444,37 @@ class WindowModel(BaseWindowModel):
     # Hooks for WM
     #########################################
 
-    def show(self):
+    def show(self) -> None:
         self.map()
         self._internal_set_property("shown", True)
         if self.get_property("iconic"):
             self.set_property("iconic", False)
         self._update_client_geometry()
-        self.corral_window.show_unraised()
+        with xlog:
+            if not X11Window.is_mapped(self.corral_xid):
+                X11Window.MapWindow(self.corral_xid)
 
-    def hide(self):
+    def hide(self) -> None:
         self._internal_set_property("shown", False)
-        self.corral_window.hide()
-        self.corral_window.reparent(self.parking_window, 0, 0)
-        self.send_configure_notify()
-
-    def send_configure_notify(self):
-        with xswallow:
+        with xlog:
+            if X11Window.is_mapped(self.corral_xid):
+                X11Window.Unmap(self.corral_xid)
+            if X11Window.getParent(self.corral_xid) != self.parking_window_xid:
+                X11Window.Reparent(self.corral_xid, self.parking_window_xid, 0, 0)
             X11Window.sendConfigureNotify(self.xid)
 
 
-    def _update_client_geometry(self):
+    def _update_client_geometry(self) -> None:
         """ figure out where we're supposed to get the window geometry from,
             and call do_update_client_geometry which will send a Configure and Notify
         """
         geometry = self.get_property("client-geometry")
         if geometry is not None:
             geomlog("_update_client_geometry: using client-geometry=%s", geometry)
+            x, y, w, h = geometry
+            self._internal_set_property("set-initial-position", True)
+            self._internal_set_property("requested-position", (x, y))
+            self._internal_set_property("requested-size", (w, h))
         elif not self._setup_done:
             #try to honour initial size and position requests during setup:
             w, h = self.get_property("requested-size")
@@ -484,54 +487,54 @@ class WindowModel(BaseWindowModel):
         self._do_update_client_geometry(geometry)
 
 
-    def _do_update_client_geometry(self, geometry):
+    def _do_update_client_geometry(self, geometry) -> None:
         geomlog("_do_update_client_geometry(%s)", geometry)
         hints = self.get_property("size-hints")
         x, y, allocated_w, allocated_h = geometry
         w, h = self.calc_constrained_size(allocated_w, allocated_h, hints)
         geomlog("_do_update_client_geometry: size(%s)=%ix%i", hints, w, h)
-        self.corral_window.move_resize(x, y, w, h)
-        self._updateprop("geometry", (x, y, w, h))
         with xlog:
+            X11Window.MoveResizeWindow(self.corral_xid, x, y, w, h)
+            self._updateprop("geometry", (x, y, w, h))
             X11Window.configureAndNotify(self.xid, 0, 0, w, h)
 
-    def do_xpra_configure_event(self, event):
-        cxid = self.corral_window.get_xid()
+    def do_xpra_configure_event(self, event) -> None:
         geomlog("WindowModel.do_xpra_configure_event(%s) corral=%#x, client=%#x, managed=%s",
-                event, cxid, self.xid, self._managed)
-        if not self._managed:
+                event, self.corral_xid, self.xid, self._managed)
+        if not self._managed or not self.corral_xid or not self.xid:
             return
-        if event.window==self.corral_window:
+        if event.window==self.corral_xid:
             #we only care about events on the client window
-            geomlog("WindowModel.do_xpra_configure_event: event is on the corral window %#x, ignored", cxid)
+            geomlog(f"ignored configure event on the corral window {self.corral_xid:x}")
             return
-        if event.window!=self.client_window:
+        if event.window!=self.xid:
             #we only care about events on the client window
-            geomlog("WindowModel.do_xpra_configure_event: event is not on the client window but on %#x, ignored",
-                    event.window.get_xid())
-            return
-        if self.corral_window is None or not self.corral_window.is_visible():
-            geomlog("WindowModel.do_xpra_configure_event: corral window is not visible")
-            return
-        if self.client_window is None or not self.client_window.is_visible():
-            geomlog("WindowModel.do_xpra_configure_event: client window is not visible")
+            geomlog(f"ignored configure event on window {event.window:x} (not the client window)")
             return
         try:
             #workaround applications whose windows disappear from underneath us:
             with xsync:
+                if not X11Window.is_mapped(self.corral_xid):
+                    geomlog(f"WindowModel.do_xpra_configure_event: corral window {self.corral_xid:x} is not visible")
+                    return
+                if not X11Window.is_mapped(self.xid):
+                    geomlog(f"WindowModel.do_xpra_configure_event: client window {self.xid:x} is not visible")
+                    return
                 #event.border_width unused
                 self.resize_corral_window(event.x, event.y, event.width, event.height)
                 self.update_children()
         except XError as e:
             geomlog("do_xpra_configure_event(%s)", event, exc_info=True)
-            geomlog.warn("Warning: failed to resize corral window %#x", cxid)
+            geomlog.warn("Warning: failed to resize corral window %#x", self.corral_xid)
             geomlog.warn(" %s", e)
 
-    def update_children(self):
-        ww, wh = self.client_window.get_geometry()[2:4]
+    def update_children(self) -> None:
+        geom = X11Window.getGeometry(self.xid)
+        if not geom:
+            return
+        ww, wh = geom[2:4]
         children = []
-        for w in get_children(self.client_window):
-            xid = w.get_xid()
+        for xid in X11Window.get_children(self.xid):
             if X11Window.is_inputonly(xid):
                 continue
             geom = X11Window.getGeometry(xid)
@@ -547,10 +550,10 @@ class WindowModel(BaseWindowModel):
             children.append([xid]+list(geom))
         self._internal_set_property("children", children)
 
-    def resize_corral_window(self, x : int, y : int, w : int, h : int):
+    def resize_corral_window(self, x : int, y : int, w : int, h : int) -> None:
         #the client window may have been resized or moved (generally programmatically)
         #so we may need to update the corral_window to match
-        cox, coy, cow, coh = self.corral_window.get_geometry()[:4]
+        cow, coh = X11Window.getGeometry(self.corral_xid)[2:4]
         #size changes (and position if any):
         hints = self.get_property("size-hints")
         w, h = self.calc_constrained_size(w, h, hints)
@@ -559,36 +562,36 @@ class WindowModel(BaseWindowModel):
         moved = x!=0 or y!=0
         geomlog("resize_corral_window%s hints=%s, constrained size=%s, geometry=%s, resized=%s, moved=%s",
                 (x, y, w, h), hints, (w, h), (cx, cy, cw, ch), resized, moved)
+        if not (moved or resized):
+            return
         if moved:
             self._internal_set_property("requested-position", (x, y))
             self._internal_set_property("set-initial-position", True)
-
         if resized:
-            if moved:
-                geomlog("resize_corral_window() move and resize from %s to %s", (cox, coy, cow, coh), (x, y, w, h))
-                self.corral_window.move_resize(x, y, w, h)
-                self.client_window.move(0, 0)
-                self._updateprop("geometry", (x, y, w, h))
-            else:
-                geomlog("resize_corral_window() resize from %s to %s", (cow, coh), (w, h))
-                self.corral_window.resize(w, h)
-                self._updateprop("geometry", (cx, cy, w, h))
+            self._internal_set_property("requested-size", (w, h))
+        if moved and resized:
+            X11Window.MoveResizeWindow(self.corral_xid, x, y, w, h)
+            X11Window.MoveWindow(self.xid, 0, 0)
         elif moved:
-            geomlog("resize_corral_window() moving corral window from %s to %s", (cox, coy), (x, y))
-            self.corral_window.move(x, y)
-            self.client_window.move(0, 0)
-            self._updateprop("geometry", (x, y, cw, ch))
+            # ICCCM 4.2.3
+            X11Window.MoveWindow(self.corral_xid, x, y)
+            X11Window.MoveWindow(self.xid, 0, 0)
+            self.send_configure_notify()
+        elif resized:
+            X11Window.ResizeWindow(self.corral_xid, w, h)
+            x = cx
+            y = cy
+        self._updateprop("geometry", (x, y, w, h))
 
-    def do_child_configure_request_event(self, event):
-        cxid = self.corral_window.get_xid()
+    def do_child_configure_request_event(self, event) -> None:
         hints = self.get_property("size-hints")
         geomlog("do_child_configure_request_event(%s) client=%#x, corral=%#x, value_mask=%s, size-hints=%s",
-                event, self.xid, cxid, configure_bits(event.value_mask), hints)
+                event, self.xid, self.corral_xid, configure_bits(event.value_mask), hints)
         if event.value_mask & CWStackMode:
             geomlog(" restack above=%s, detail=%s", event.above, event.detail)
         # Also potentially update our record of what the app has requested:
-        ogeom = self.get_property("geometry")
-        x, y, w, h = ogeom[:4]
+        ox, oy, ow, oh = self.get_property("geometry")
+        x, y, w, h = ox, oy, ow, oh
         rx, ry = self.get_property("requested-position")
         if event.value_mask & CWX:
             x = event.x
@@ -615,51 +618,67 @@ class WindowModel(BaseWindowModel):
 
         if VALIDATE_CONFIGURE_REQUEST:
             w, h = self.calc_constrained_size(w, h, hints)
-        #update the geometry now, as another request may come in
-        #before we've had a chance to process the ConfigureNotify that the code below will generate
-        self._updateprop("geometry", (x, y, w, h))
-        geomlog("do_child_configure_request_event updated requested geometry from %s to  %s", ogeom, (x, y, w, h))
-        # As per ICCCM 4.1.5, even if we ignore the request
-        # send back a synthetic ConfigureNotify telling the client that nothing has happened.
-        with xswallow:
-            X11Window.configureAndNotify(self.xid, x, y, w, h, event.value_mask)
+
+        # update the geometry now, as another request may come in
+        # before we've had a chance to process the ConfigureNotify that the code below will generate
+        if not self._updateprop("geometry", (x, y, w, h)):
+            # As per ICCCM 4.1.5, even if we ignore the request
+            # send back a synthetic ConfigureNotify telling the client that nothing has happened.
+            geomlog("geometry unchanged: %s", (x, y, w, h))
+            with xlog:
+                X11Window.sendConfigureNotify(self.xid)
+            return
+
+        geomlog("do_child_configure_request_event updated requested geometry from %s to %s",
+                (ox, oy, ow, oh), (x, y, w, h))
+        mask = 0
+        if ox != x:
+            mask |= CWX
+        if oy != y:
+            mask |= CWY
+        if ow != w:
+            mask |= CWWidth
+        if oh != h:
+            mask |= CWHeight
+        with xlog:
+            X11Window.configure(self.xid, x, y, w, h, mask)
         # FIXME: consider handling attempts to change stacking order here.
         # (In particular, I believe that a request to jump to the top is
         # meaningful and should perhaps even be respected.)
 
-    def process_client_message_event(self, event):
+    def process_client_message_event(self, event) -> bool:
         if event.message_type=="_NET_MOVERESIZE_WINDOW":
             #TODO: honour gravity, show source indication
-            geom = self.corral_window.get_geometry()
-            x, y, w, h, _ = geom
-            if event.data[0] & 0x100:
-                x = event.data[1]
-            if event.data[0] & 0x200:
-                y = event.data[2]
-            if event.data[0] & 0x400:
-                w = event.data[3]
-            if event.data[0] & 0x800:
-                h = event.data[4]
-            if (event.data[0] & 0x100) or (event.data[0] & 0x200):
-                self._internal_set_property("set-initial-position", True)
-                self._internal_set_property("requested-position", (x, y))
-            #honour hints:
-            hints = self.get_property("size-hints")
-            w, h = self.calc_constrained_size(w, h, hints)
-            geomlog("_NET_MOVERESIZE_WINDOW on %s (data=%s, current geometry=%s, new geometry=%s)",
-                    self, event.data, geom, (x,y,w,h))
-            with xswallow:
+            with xsync:
+                geom = X11Window.getGeometry(self.corral_xid)
+                x, y, w, h, _ = geom
+                if event.data[0] & 0x100:
+                    x = event.data[1]
+                if event.data[0] & 0x200:
+                    y = event.data[2]
+                if event.data[0] & 0x400:
+                    w = event.data[3]
+                if event.data[0] & 0x800:
+                    h = event.data[4]
+                if (event.data[0] & 0x100) or (event.data[0] & 0x200):
+                    self._internal_set_property("set-initial-position", True)
+                    self._internal_set_property("requested-position", (x, y))
+                #honour hints:
+                hints = self.get_property("size-hints")
+                w, h = self.calc_constrained_size(w, h, hints)
+                geomlog("_NET_MOVERESIZE_WINDOW on %s (data=%s, current geometry=%s, new geometry=%s)",
+                        self, event.data, geom, (x,y,w,h))
                 X11Window.configureAndNotify(self.xid, x, y, w, h)
             return True
         return super().process_client_message_event(event)
 
-    def calc_constrained_size(self, w, h, hints):
+    def calc_constrained_size(self, w, h, hints) -> Tuple[int,int]:
         mhints = typedict(hints)
         cw, ch = calc_constrained_size(w, h, mhints)
         geomlog("calc_constrained_size%s=%s (size_constraints=%s)", (w, h, mhints), (cw, ch), self.size_constraints)
         return cw, ch
 
-    def update_size_constraints(self, minw=0, minh=0, maxw=MAX_WINDOW_SIZE, maxh=MAX_WINDOW_SIZE):
+    def update_size_constraints(self, minw:int=0, minh:int=0, maxw:int=MAX_WINDOW_SIZE, maxh:int=MAX_WINDOW_SIZE) -> None:
         if self.size_constraints==(minw, minh, maxw, maxh):
             geomlog("update_size_constraints%s unchanged", (minw, minh, maxw, maxh))
             return  #no need to do anything
@@ -676,7 +695,7 @@ class WindowModel(BaseWindowModel):
     # X11 properties synced to Python objects
     #########################################
 
-    def _handle_icon_title_change(self):
+    def _handle_icon_title_change(self) -> None:
         icon_name = self.prop_get("_NET_WM_ICON_NAME", "utf8", True)
         iconlog("_NET_WM_ICON_NAME=%s", icon_name)
         if icon_name is None:
@@ -684,10 +703,9 @@ class WindowModel(BaseWindowModel):
             iconlog("WM_ICON_NAME=%s", icon_name)
         self._updateprop("icon-title", sanestr(icon_name))
 
-    def _handle_motif_wm_hints_change(self):
+    def _handle_motif_wm_hints_change(self) -> None:
         #motif_hints = self.prop_get("_MOTIF_WM_HINTS", "motif-hints")
-        motif_hints = prop_get(self.client_window, "_MOTIF_WM_HINTS", "motif-hints",
-                               ignore_errors=False, raise_xerrors=True)
+        motif_hints = self.prop_get("_MOTIF_WM_HINTS", "motif-hints", ignore_errors=False, raise_xerrors=True)
         metalog("_MOTIF_WM_HINTS=%s", motif_hints)
         if motif_hints:
             if motif_hints.flags & (2**MotifWMHints.DECORATIONS_BIT):
@@ -698,7 +716,7 @@ class WindowModel(BaseWindowModel):
                 self._updateprop("modal", int(motif_hints.input_mode))
 
 
-    def _handle_wm_normal_hints_change(self):
+    def _handle_wm_normal_hints_change(self) -> None:
         with xswallow:
             size_hints = X11Window.getSizeHints(self.xid)
         metalog("WM_NORMAL_HINTS=%s", size_hints)
@@ -725,7 +743,7 @@ class WindowModel(BaseWindowModel):
         mhints = typedict(size_hints or {})
         hminw, hminh = mhints.inttupleget("min_size", (0, 0), 2, 2)
         hmaxw, hmaxh = mhints.inttupleget("max_size", (MAX_WINDOW_SIZE, MAX_WINDOW_SIZE), 2, 2)
-        d = self.get("decorations", -1)
+        d : int = self.get("decorations", -1)
         decorated = d==-1 or any((d & 2**b) for b in (
             MotifWMHints.ALL_BIT,
             MotifWMHints.TITLE_BIT,
@@ -763,7 +781,7 @@ class WindowModel(BaseWindowModel):
                 self._update_client_geometry()
 
 
-    def _handle_net_wm_icon_change(self):
+    def _handle_net_wm_icon_change(self) -> None:
         iconlog("_NET_WM_ICON changed on %#x, re-reading", self.xid)
         icons = self.prop_get("_NET_WM_ICON", "icons")
         self._internal_set_property("icons", icons)
@@ -778,44 +796,12 @@ class WindowModel(BaseWindowModel):
        })
 
 
-    def get_default_window_icon(self, size=48):
-        #return the icon which would be used from the wmclass
-        c_i = self.get_property("class-instance")
-        iconlog("get_default_window_icon(%i) class-instance=%s", size, c_i)
-        if not c_i or len(c_i)!=2:
-            return None
-        wmclass_name = c_i[0]
-        if not wmclass_name:
-            return None
-        it = Gtk.IconTheme.get_default()  # pylint: disable=no-member
-        pixbuf = None
-        iconlog("get_default_window_icon(%i) icon theme=%s, wmclass_name=%s", size, it, wmclass_name)
-        for icon_name in (
-            f"{wmclass_name}-color",
-            wmclass_name,
-            f"{wmclass_name}_{size}x{size}",
-            f"application-x-{wmclass_name}",
-            f"{wmclass_name}-symbolic",
-            f"{wmclass_name}.symbolic",
-            ):
-            i = it.lookup_icon(icon_name, size, 0)
-            iconlog("lookup_icon(%s)=%s", icon_name, i)
-            if not i:
-                continue
-            try:
-                pixbuf = i.load_icon()
-                iconlog("load_icon()=%s", pixbuf)
-                if pixbuf:
-                    w, h = pixbuf.props.width, pixbuf.props.height
-                    iconlog("using '%s' pixbuf %ix%i", icon_name, w, h)
-                    return w, h, "RGBA", pixbuf.get_pixels()
-            except Exception:
-                iconlog("%s.load_icon()", i, exc_info=True)
+    def get_default_window_icon(self, size:int=48):
         return None
 
-    def get_wm_state(self, prop):
+    def get_wm_state(self, prop) -> bool:
         state_names = self._state_properties.get(prop)
-        assert state_names, "invalid window state %s" % prop
+        assert state_names, f"invalid window state {prop}"
         log("get_wm_state(%s) state_names=%s", prop, state_names)
         #this is a virtual property for _NET_WM_STATE:
         #return True if any is set (only relevant for maximized)
@@ -829,15 +815,15 @@ class WindowModel(BaseWindowModel):
     # Focus handling:
     ################################
 
-    def give_client_focus(self):
+    def give_client_focus(self) -> None:
         """The focus manager has decided that our client should receive X
         focus.  See world_window.py for details."""
-        log("give_client_focus() corral_window=%s", self.corral_window)
-        if self.corral_window:
+        log("give_client_focus() corral_xid=%s", self.corral_xid)
+        if self.corral_xid:
             with xlog:
                 self.do_give_client_focus()
 
-    def do_give_client_focus(self):
+    def do_give_client_focus(self) -> None:
         protocols = self.get_property("protocols")
         focuslog("Giving focus to %#x, input_field=%s, FORCE_XSETINPUTFOCUS=%s, protocols=%s",
                  self.xid, self._input_field, FORCE_XSETINPUTFOCUS, protocols)
@@ -846,7 +832,7 @@ class WindowModel(BaseWindowModel):
         # genuine race conditions here (e.g. suppose the client does not
         # actually get around to requesting the focus until after we have
         # already changed our mind and decided to give it to someone else).
-        now = x11_get_server_time(self.corral_window)
+        now = X11Window.get_server_time(self.corral_xid)
         # ICCCM 4.1.7 *claims* to describe how we are supposed to give focus
         # to a window, but it is completely opaque.  From reading the
         # metacity, kwin, gtk+, and qt code, it appears that the actual rules
@@ -856,7 +842,7 @@ class WindowModel(BaseWindowModel):
         #   -- independently, the WM_TAKE_FOCUS protocol determines whether
         #      the WM should send a WM_TAKE_FOCUS ClientMessage.
         # If both are set, both methods MUST be used together. For example,
-        # GTK+ apps respect WM_TAKE_FOCUS alone but I'm not sure they handle
+        # GTK+ apps respect WM_TAKE_FOCUS alone, but I'm not sure they handle
         # XSetInputFocus well, while Qt apps ignore (!!!) WM_TAKE_FOCUS
         # (unless they have a modal window), and just expect to get focus from
         # the WM's XSetInputFocus.
@@ -865,7 +851,7 @@ class WindowModel(BaseWindowModel):
             X11Window.XSetInputFocus(self.xid, now)
         if "WM_TAKE_FOCUS" in protocols:
             focuslog("... using WM_TAKE_FOCUS")
-            send_wm_take_focus(self.client_window, now)
+            send_wm_take_focus(self.xid, now)
         self.set_active()
 
 
